import numpy as np

from scipy import special


__all__ = ['FiniteStateMachine']


class FiniteStateMachine:
    """
    Finite-state machine (Mealy machine). It is defined by a *set of states* :math:`\\mathcal{S}`, an *input alphabet* :math:`\\mathcal{X}`, an *output alphabet* :math:`\\mathcal{Y}`, and a *transition function* :math:`T : \\mathcal{S} \\times \\mathcal{X} \\to \\mathcal{S} \\times \\mathcal{Y}`. Here, for simplicity, the set of states, the input alphabet, and the output alphabet are always taken as :math:`\\mathcal{S} = \\{ 0, 1, \\ldots, |\\mathcal{S}| - 1 \\}`, :math:`\\mathcal{X} = \\{ 0, 1, \\ldots, |\\mathcal{X}| - 1 \\}`, and :math:`\\mathcal{Y} = \\{ 0, 1, \\ldots, |\\mathcal{Y}| - 1 \\}`, respectively.

    For example, consider the finite-state machine whose state diagram depicted in the figure below.

    .. image:: figures/mealy.png
       :alt: Finite-state machine (Mealy machine) example.
       :align: center

    It has set of states :math:`\\mathcal{S} = \\{ 0, 1, 2, 3 \\}`, input alphabet :math:`\\mathcal{X} = \\{ 0, 1 \\}`, output alphabet :math:`\\mathcal{Y} = \\{ 0, 1, 2, 3 \\}`, and transition function :math:`T` given by the table below.

    .. csv-table:: Transition function
       :align: center
       :header: State, Input, Next state, Output

       0, 0, 0, 0
       0, 1, 1, 3
       1, 0, 2, 1
       1, 1, 3, 2
       2, 0, 0, 3
       2, 1, 1, 0
       3, 0, 2, 2
       3, 1, 3, 1

    |
    """
    def __init__(self, next_states, outputs):
        """
        Constructor for the class. It expects the following parameters:

        :code:`next_states` : 2D-array of :obj:`int`
            The matrix of next states of the machine, of shape :math:`|\\mathcal{S}| \\times |\\mathcal{X}|`. The element in row :math:`s` and column :math:`x` should be the next state of the machine (an element in :math:`\\mathcal{S}`), given that the current state is :math:`s \\in \\mathcal{S}` and the input is :math:`x \\in \\mathcal{X}`.

        :code:`outputs` : 2D-array of :obj:`int`
            The matrix of outputs of the machine, of shape :math:`|\\mathcal{S}| \\times |\\mathcal{X}|`. The element in row :math:`s` and column :math:`x` should be the output of the machine (an element in :math:`\\mathcal{Y}`), given that the current state is :math:`s \\in \\mathcal{S}` and the input is :math:`x \\in \\mathcal{X}`.

        .. rubric:: Examples

        >>> fsm = komm.FiniteStateMachine(next_states=[[0,1], [2,3], [0,1], [2,3]], outputs=[[0,3], [1,2], [3,0], [2,1]])
        """
        self._next_states = np.array(next_states, dtype=np.int)
        self._outputs = np.array(outputs, dtype=np.int)
        self._num_states, self._num_input_symbols = self._next_states.shape
        self._num_output_symbols = np.amax(self._outputs)

        self._input_edges = np.full((self._num_states, self._num_states), fill_value=-1)
        self._output_edges = np.full((self._num_states, self._num_states), fill_value=-1)
        for state_from in range(self._num_states):
            for x, state_to in enumerate(self._next_states[state_from, :]):
                self._input_edges[state_from, state_to] = x
                self._output_edges[state_from, state_to] = self._outputs[state_from, x]

    def __repr__(self):
        args = 'next_states={}, outputs={}'.format(self._next_states.tolist(), self._outputs.tolist())
        return '{}({})'.format(self.__class__.__name__, args)

    @property
    def num_states(self):
        """
        The number of states of the machine. This property is read-only.
        """
        return self._num_states

    @property
    def num_input_symbols(self):
        """
        The size (cardinality) of the input alphabet :math:`\\mathcal{X}`. This property is read-only.
        """
        return self._num_input_symbols

    @property
    def num_output_symbols(self):
        """
        The size (cardinality) of the output alphabet :math:`\\mathcal{Y}`. This property is read-only.
        """
        return self._num_output_symbols

    @property
    def next_states(self):
        """
        The matrix of next states of the machine. It has shape :math:`|\\mathcal{S}| \\times |\\mathcal{X}|`. The element in row :math:`s` and column :math:`x` is the next state of the machine (an element in :math:`\\mathcal{S}`), given that the current state is :math:`s \\in \\mathcal{S}` and the input is :math:`x \\in \\mathcal{X}`. This property is read-only.
        """
        return self._next_states

    @property
    def outputs(self):
        """
        The matrix of outputs of the machine. It has shape :math:`|\\mathcal{S}| \\times |\\mathcal{X}|`. The element in row :math:`s` and column :math:`x` is the output of the machine (an element in :math:`\\mathcal{Y}`), given that the current state is :math:`s \\in \\mathcal{S}` and the input is :math:`x \\in \\mathcal{X}`. This property is read-only.
        """
        return self._outputs

    @property
    def input_edges(self):
        """
        The matrix of input edges of the machine. It has shape :math:`|\\mathcal{S}| \\times |\\mathcal{S}|`. If there is an edge from :math:`s_0 \\in \\mathcal{S}` to :math:`s_1 \\in \\mathcal{S}`, then the element in row :math:`s_0` and column :math:`s_1` is the input associated with that edge (an element of :math:`\\mathcal{X}`); if there is no such edge, then the element is :math:`-1`. This property is read-only.

        .. rubric:: Example

        >>> fsm = komm.FiniteStateMachine(next_states=[[0,1], [2,3], [0,1], [2,3]], outputs=[[0,3], [1,2], [3,0], [2,1]])
        >>> fsm.input_edges
        array([[ 0,  1, -1, -1],
               [-1, -1,  0,  1],
               [ 0,  1, -1, -1],
               [-1, -1,  0,  1]])
        """
        return self._input_edges

    @property
    def output_edges(self):
        """
        The matrix of output edges of the machine. It has shape :math:`|\\mathcal{S}| \\times |\\mathcal{S}|`. If there is an edge from :math:`s_0 \\in \\mathcal{S}` to :math:`s_1 \\in \\mathcal{S}`, then the element in row :math:`s_0` and column :math:`s_1` is the output associated with that edge (an element of :math:`\\mathcal{Y}`); if there is no such edge, then the element is :math:`-1`. This property is read-only.

        .. rubric:: Example

        >>> fsm = komm.FiniteStateMachine(next_states=[[0,1], [2,3], [0,1], [2,3]], outputs=[[0,3], [1,2], [3,0], [2,1]])
        >>> fsm.output_edges
        array([[ 0,  3, -1, -1],
               [-1, -1,  1,  2],
               [ 3,  0, -1, -1],
               [-1, -1,  2,  1]])
        """
        return self._output_edges

    def process(self, input_sequence, initial_state):
        """
        Returns the output sequence corresponding to a given input sequence. It assumes the machine starts at a given initial state :math:`s_\\mathrm{i}`. The input sequence and the output sequence are denoted by :math:`\\mathbf{x} = (x_0, x_1, \\ldots, x_{L-1}) \\in \\mathcal{X}^L` and :math:`\\mathbf{y} = (y_0, y_1, \\ldots, y_{L-1}) \\in \\mathcal{Y}^L`, respectively.

        **Input:**

        :code:`input_sequence` : 1D-array of :obj:`int`
            The input sequence :math:`\\mathbf{x} \\in \\mathcal{X}^L`. It should be a 1D-array with elements in :math:`\\mathcal{X}`.

        :code:`initial_state` : :obj:`int`
            The initial state :math:`s_\\mathrm{i}` of the machine. Should be an integer in :math:`\\mathcal{S}`.

        **Output:**

        :code:`output_sequence` : 1D-array of :obj:`int`
            The output sequence :math:`\\mathbf{y} \\in \\mathcal{Y}^L` corresponding to :code:`input_sequence`, assuming the machine starts at the state given by :code:`initial_state`. It is a 1D-array with elements in :math:`\\mathcal{Y}`.

        :code:`final_state` : :obj:`int`
            The final state :math:`s_\\mathrm{f}` of the machine. It is an integer in :math:`\\mathcal{S}`.

        .. rubric:: Example

        >>> fsm = komm.FiniteStateMachine(next_states=[[0,1], [2,3], [0,1], [2,3]], outputs=[[0,3], [1,2], [3,0], [2,1]])
        >>> input_sequence, initial_state = [1, 1, 0, 1, 0], 0
        >>> output_sequence, final_state = fsm.process(input_sequence, initial_state)
        >>> output_sequence
        array([3, 2, 2, 0, 1])
        >>> final_state
        2
        """
        output_sequence = np.empty_like(input_sequence, dtype=np.int)
        s = initial_state
        for t, x in np.ndenumerate(input_sequence):
            y = self._outputs[s, x]
            s = self._next_states[s, x]
            output_sequence[t] = y
        final_state = s
        return output_sequence, final_state

    def viterbi(self, observed_sequence, metric_function, initial_metrics=None):
        """
        Applies the Viterbi algorithm on a given observed sequence. The Viterbi algorithm finds the most probable input sequence :math:`\\hat{\\mathbf{x}}(s) \\in \\mathcal{X}^L` ending in state :math:`s`, for all :math:`s \\in \\mathcal{S}`, given an observed sequence :math:`\\mathbf{z} \\in \\mathcal{Z}^L`. It is assumed uniform input priors.

        References: :cite:`Lin.Costello.04` (Sec. 12.1).

        **Input:**

        :code:`observed_sequence` : 1D-array
            The observed sequence :math:`\\mathbf{z} \\in \\mathcal{Z}^L`.

        :code:`metric_function` : function
            The metric function :math:`\\mathcal{Y} \\times \\mathcal{Z} \\to \\mathbb{R}`.

        :code:`initial_metrics` : 1D-array of :obj:`float`, optional
            The initial metrics for each state. It must be a 1D-array of length :math:`|\\mathcal{S}|`. The default value is :code:`0.0` for all states.

        **Output:**

        :code:`input_sequences_hat` : 2D-array of :obj:`int`
            The most probable input sequence :math:`\\hat{\\mathbf{x}}(s) \\in \\mathcal{X}^L` ending in state :math:`s`, for all :math:`s \\in \\mathcal{S}`. It is a 2D-array of shape :math:`L \\times |\\mathcal{S}|`, in which column :math:`s` is equal to :math:`\\hat{\\mathbf{x}}(s)`.

        :code:`final_metrics` : 1D-array of :obj:`float`
            The final metrics for each state. It is a 1D-array of length :math:`|\\mathcal{S}|`.
        """
        L, num_states = len(observed_sequence), self._num_states
        choices = np.empty((L, num_states), dtype=np.int)
        metrics = np.full((L + 1, num_states), fill_value=np.inf)
        if initial_metrics is None:
            metrics[0, :] = np.zeros(num_states, dtype=np.float)
        else:
            metrics[0, :] = initial_metrics
        for t, z in enumerate(observed_sequence):
            for s0 in range(num_states):
                for (s1, y) in zip(self._next_states[s0], self._outputs[s0]):
                    candidate_metrics = metrics[t, s0] + metric_function(y, z)
                    if candidate_metrics < metrics[t + 1, s1]:
                        metrics[t + 1, s1] = candidate_metrics
                        choices[t, s1] = s0

        # Backtrack
        input_sequences_hat = np.empty((L, num_states), dtype=np.int)
        for final_state in range(num_states):
            s1 = final_state
            for t in reversed(range(L)):
                s0 = choices[t, s1]
                input_sequences_hat[t, final_state] = self._input_edges[s0, s1]
                s1 = s0

        return input_sequences_hat, metrics[L, :]

    def viterbi_streaming(self, observed_sequence, metric_function, memory):
        """
        Applies the streaming version of the Viterbi algorithm on a given observed sequence. The path memory (or traceback length) is denoted by :math:`\\tau`. It chooses the survivor with best metric and selects the information block on this path.

        References: :cite:`Lin.Costello.04` (Sec. 12.3).

        **Input:**

        :code:`observed_sequence` : 1D-array
            The observed sequence :math:`\\mathbf{z} \\in \\mathcal{Z}^L`.

        :code:`metric_function` : function
            The metric function :math:`\\mathcal{Y} \\times \\mathcal{Z} \\to \\mathbb{R}`.

        **Output:**

        :code:`input_sequence_hat` : 1D-array of :obj:`int`
            The most probable input sequence. It is a 1D-array of length :math:`L`.

        **Input and output:**

        :code:`memory` : :obj:`dict`
            The past metrics for each state. It must be a 2D-array of shape :math:`|\\mathcal{S}| \\times (\\tau + 1)`.
            The initial metrics for each state. It must be a 2D-array of shape :math:`|\\mathcal{S}| \\times (\\tau + 1)`.
        """
        num_states = self._num_states
        input_sequences_hat = np.empty(len(observed_sequence), dtype=np.int)
        for t, z in enumerate(observed_sequence):
            new_metrics = np.full(num_states, fill_value=np.inf)
            choices = np.zeros(num_states, dtype=np.int)
            for s0 in range(num_states):
                for (s1, y) in zip(self._next_states[s0], self._outputs[s0]):
                    candidate_metric = memory['metrics'][s0, -1] + metric_function(y, z)
                    if candidate_metric < new_metrics[s1]:
                        new_metrics[s1] = candidate_metric
                        choices[s1] = s0

            s_star = np.argmin(new_metrics)
            s0, s1 = memory['paths'][s_star, :2]
            input_sequences_hat[t] = self._input_edges[s0, s1]

            memory['metrics'] = np.roll(memory['metrics'], shift=-1, axis=1)
            memory['metrics'][:, -1] = new_metrics
            memory['paths'] = np.roll(memory['paths'], shift=-1, axis=1)

            paths_copy = np.copy(memory['paths'])
            for s1, s0 in enumerate(choices):
                memory['paths'][s1, :-1] = paths_copy[s0, :-1]
                memory['paths'][s1, -1] = s1

        return input_sequences_hat

    def forward_backward(self, observed_sequence, metric_function, input_priors=None, initial_state_distribution=None, final_state_distribution=None):
        """
        Applies the forward-backward algorithm on a given observed sequence. The forward-backward algorithm computes the posterior :term:`pmf` of each input :math:`x_0, x_1, \\ldots, x_{L-1} \\in \\mathcal{X}` given an observed sequence :math:`\\mathbf{z} = (z_0, z_1, \\ldots, z_{L-1}) \\in \\mathcal{Z}^L`. The prior :term:`pmf` of each input may also be provided.

        References: :cite:`Lin.Costello.04` (Sec. 12.6).

        **Input:**

        :code:`observed_sequence` : 1D-array
            The observed sequence :math:`\\mathbf{z} \\in \\mathcal{Z}^L`.

        :code:`metric_function` : function
            The metric function :math:`\\mathcal{Y} \\times \\mathcal{Z} \\to \\mathbb{R}`.

        :code:`input_priors` : 2D-array of :obj:`float`, optional
            The prior :term:`pmf` of each input, of shape :math:`L \\times |\\mathcal{X}|`. The element in row :math:`t \\in [0 : L)` and column :math:`x \\in \\mathcal{X}` should be :math:`p(x_t = x)`. The default value considers uniform priors.

        :code:`initial_state_distribution` : 1D-array of :obj:`float`, optional
            The :term:`pmf` of the initial state of the machine. It must be a 1D-array of length :math:`|\\mathcal{S}|`. The default value is uniform over all states.

        :code:`final_state_distribution` : 1D-array of :obj:`float`, optional
            The :term:`pmf` of the final state of the machine. It must be a 1D-array of length :math:`|\\mathcal{S}|`. The default value is uniform over all states.

        **Output:**

        :code:`input_posteriors` : 2D-array of :obj:`float`
            The posterior :term:`pmf` of each input, given the observed sequence, of shape :math:`L \\times |\\mathcal{X}|`. The element in row :math:`t \\in [0 : L)` and column :math:`x \\in \\mathcal{X}` is :math:`p(x_t = x \\mid \\mathbf{z})`.
        """
        L, num_states, num_input_symbols = len(observed_sequence), self._num_states, self._num_input_symbols

        if input_priors is None:
            input_priors = np.ones((L, num_input_symbols)) / num_input_symbols
        if initial_state_distribution is None:
            initial_state_distribution = np.ones(num_states) / num_states
        if initial_state_distribution is None:
            final_state_distribution = np.ones(num_states) / num_states

        log_gamma = np.full((L, num_states, num_states), fill_value=-np.inf)
        log_alpha = np.full((L + 1, num_states), fill_value=-np.inf)
        log_beta = np.full((L + 1, num_states), fill_value=-np.inf)

        with np.errstate(divide='ignore'):
            log_input_priors = np.log(input_priors)
            log_alpha[0, :] = np.log(initial_state_distribution)
            log_beta[L, :] = np.log(final_state_distribution)

        for t, z in enumerate(observed_sequence):
            for x, s0 in np.ndindex(num_input_symbols, num_states):
                y, s1 = self._outputs[s0, x], self._next_states[s0, x]
                log_gamma[t, s0, s1] = log_input_priors[t, x] + metric_function(y, z)

        for t in range(0, L - 1):
            for s1 in range(num_states):
                log_alpha[t + 1, s1] = special.logsumexp(log_gamma[t, :, s1] + log_alpha[t, :])

        for t in range(L - 1, -1, -1):
            for s0 in range(num_states):
                log_beta[t, s0] = special.logsumexp(log_gamma[t, s0, :] + log_beta[t + 1, :])

        log_input_posteriors = np.empty((L, num_input_symbols), dtype=np.float)
        edge_labels = np.empty(num_states, dtype=np.float)
        for t in range(L):
            for x in range(num_input_symbols):
                for s0 in range(num_states):
                    s1 = self._next_states[s0, x]
                    edge_labels[s0] = log_alpha[t, s0] + log_gamma[t, s0, s1] + log_beta[t + 1, s1]
                log_input_posteriors[t, x] = special.logsumexp(edge_labels)

        input_posteriors = np.exp(log_input_posteriors - np.amax(log_input_posteriors, axis=1, keepdims=True))
        input_posteriors /= np.sum(input_posteriors, axis=1, keepdims=True)

        return input_posteriors



import heapq
import itertools

import numpy as np

__all__ = ['FixedToVariableCode', 'VariableToFixedCode',
           'HuffmanCode', 'TunstallCode']


class FixedToVariableCode:
    """
    Binary (prefix-free) fixed-to-variable length code. Let :math:`\\mathcal{X}` be the alphabet of some discrete source. A *binary fixed-to-variable length code* of source block size :math:`k` is defined by an encoding mapping :math:`\\mathrm{Enc} : \\mathcal{X}^k \\to \\{ 0, 1 \\}^+`, where :math:`\\{ 0, 1 \\}^+` denotes the set of all finite-length, non-empty binary strings. Here, for simplicity, the source alphabet is always taken as :math:`\\mathcal{X} = \\{0, 1, \\ldots, |\\mathcal{X} - 1| \\}`. The elements in the image of :math:`\\mathrm{Enc}` are called *codewords*.

    Also, we only consider *prefix-free* codes, in which no codeword is a prefix of any other codeword.
    """
    def __init__(self, codewords, source_cardinality=None):
        """
        Constructor for the class. It expects the following parameters:

        :code:`codewords` : :obj:`list` of :obj:`tuple` of :obj:`int`
            The codewords of the code. Must be a list of length :math:`|\\mathcal{X}|^k` containing tuples of integers in :math:`\\{ 0, 1 \\}`. The tuple in position :math:`i` of :code:`codewords` should be equal to :math:`\\mathrm{Enc}(u)`, where :math:`u` is the :math:`i`-th element in the lexicographic ordering of :math:`\\mathcal{X}^k`.

        :code:`source_cardinality` : :obj:`int`, optional
            The cardinality :math:`|\\mathcal{X}|` of the source alphabet. The default value is :code:`len(codewords)`, yielding a source block size :math:`k = 1`.

        *Note:* The source block size :math:`k` is inferred from :code:`len(codewords)` and :code:`source_cardinality`.

        .. rubric:: Examples

        >>> code = komm.FixedToVariableCode(codewords=[(0,), (1,0), (1,1)])
        >>> pprint(code.enc_mapping)
        {(0,): (0,), (1,): (1, 0), (2,): (1, 1)}
        >>> pprint(code.dec_mapping)
        {(0,): (0,), (1, 0): (1,), (1, 1): (2,)}

        >>> code = komm.FixedToVariableCode(codewords=[(0,), (1,0,0), (1,1), (1,0,1)], source_cardinality=2)
        >>> pprint(code.enc_mapping)
        {(0, 0): (0,), (0, 1): (1, 0, 0), (1, 0): (1, 1), (1, 1): (1, 0, 1)}
        >>> pprint(code.dec_mapping)
        {(0,): (0, 0), (1, 0, 0): (0, 1), (1, 0, 1): (1, 1), (1, 1): (1, 0)}
        """
        # TODO: Assert prefix-free
        self._codewords = codewords
        self._source_cardinality = len(codewords) if source_cardinality is None else int(source_cardinality)
        self._source_block_size = 1
        while self._source_cardinality ** self._source_block_size < len(codewords):
            self._source_block_size += 1

        if self._source_cardinality ** self._source_block_size != len(codewords):
            raise ValueError("Invalid number of codewords")

        self._enc_mapping = {}
        self._dec_mapping = {}
        for symbols, bits in zip(itertools.product(range(self._source_cardinality), repeat=self._source_block_size), codewords):
            self._enc_mapping[symbols] = tuple(bits)
            self._dec_mapping[tuple(bits)] = symbols

    @property
    def source_cardinality(self):
        """
        The cardinality :math:`|\\mathcal{X}|` of the source alphabet.
        """
        return self._source_cardinality

    @property
    def source_block_size(self):
        """
        The source block size :math:`k`.
        """
        return self._source_block_size

    @property
    def enc_mapping(self):
        """
        The encoding mapping :math:`\\mathrm{Enc}` of the code.
        """
        return self._enc_mapping

    @property
    def dec_mapping(self):
        """
        The decoding mapping :math:`\\mathrm{Dec}` of the code.
        """
        return self._dec_mapping

    def rate(self, pmf):
        """
        Computes the expected rate :math:`R` of the code, assuming a given :term:`pmf`. It is given in bits per source symbol.

        **Input:**

        :code:`pmf` : 1D-array of :obj:`float`
            The (first-order) probability mass function :math:`p_X(x)` to be assumed.

        **Output:**

        :code:`rate` : :obj:`float`
            The expected rate :math:`R` of the code.

        .. rubric:: Examples

        >>> code = komm.FixedToVariableCode([(0,), (1,0), (1,1)])
        >>> code.rate([0.5, 0.25, 0.25])
        1.5
        """
        probabilities = np.array([np.prod(ps) for ps in itertools.product(pmf, repeat=self._source_block_size)])
        lengths = [len(bits) for bits in self._codewords]
        return np.dot(lengths, probabilities) / self._source_block_size

    def encode(self, symbol_sequence):
        """
        Encodes a given sequence of symbols to its corresponding sequence of bits.

        **Input:**

        :code:`symbol_sequence` : 1D-array of :obj:`int`
            The sequence of symbols to be encoded. Must be a 1D-array with elements in :math:`\\mathcal{X} = \\{0, 1, \\ldots, |\\mathcal{X} - 1| \\}`. Its length must be a multiple of :math:`k`.

        **Output:**

        :code:`bit_sequence` : 1D-array of :obj:`int`
            The sequence of bits corresponding to :code:`symbol_sequence`.

        .. rubric:: Examples

        >>> code = komm.FixedToVariableCode([(0,), (1,0), (1,1)])
        >>> code.encode([1, 0, 1, 0, 2, 0])
        array([1, 0, 0, 1, 0, 0, 1, 1, 0])
        """
        symbols_reshaped = np.reshape(symbol_sequence, newshape=(-1, self._source_block_size))
        return np.concatenate([self._enc_mapping[tuple(symbols)] for symbols in symbols_reshaped])

    def decode(self, bit_sequence):
        """
        Decodes a given sequence of bits to its corresponding sequence of symbols.

        **Input:**

        :code:`bit_sequence` : 1D-array of :obj:`int`
            The sequence of bits to be decoded. Must be a 1D-array with elements in :math:`\\{ 0, 1 \\}`.

        **Output:**

        :code:`symbol_sequence` : 1D-array of :obj:`int`
            The sequence of symbols corresponding to :code:`bits`.

        .. rubric:: Examples

        >>> code = komm.FixedToVariableCode([(0,), (1,0), (1,1)])
        >>> code.decode([1, 0, 0, 1, 0, 0, 1, 1, 0])
        array([1, 0, 1, 0, 2, 0])
        """
        return np.array(_parse_prefix_free(bit_sequence, self._dec_mapping))

    def __repr__(self):
        args = 'codewords={}'.format(self._codewords)
        return '{}({})'.format(self.__class__.__name__, args)


class HuffmanCode(FixedToVariableCode):
    """
    Huffman code. It is an optimal (minimal expected rate) fixed-to-variable length code (:class:`FixedToVariableCode`) for a given probability mass function.

    .. rubric:: Examples

    >>> code = komm.HuffmanCode([0.7, 0.15, 0.15])
    >>> pprint(code.enc_mapping)
    {(0,): (0,), (1,): (1, 1), (2,): (1, 0)}

    >>> code = komm.HuffmanCode([0.7, 0.15, 0.15], source_block_size=2)
    >>> pprint(code.enc_mapping)
    {(0, 0): (1,),
     (0, 1): (0, 0, 0, 0),
     (0, 2): (0, 1, 1),
     (1, 0): (0, 1, 0),
     (1, 1): (0, 0, 0, 1, 1, 1),
     (1, 2): (0, 0, 0, 1, 1, 0),
     (2, 0): (0, 0, 1),
     (2, 1): (0, 0, 0, 1, 0, 1),
     (2, 2): (0, 0, 0, 1, 0, 0)}
    """
    def __init__(self, pmf, source_block_size=1, policy='high'):
        """
        Constructor for the class. It expects the following parameters:

        :code:`pmf` : 1D-array of :obj:`float`
            The probability mass function used to construct the code.

        :code:`source_block_size` : :obj:`int`, optional
            The source block size :math:`k`. The default value is :math:`k = 1`.

        :code:`policy` : :obj:`str`, optional
            The policy to be used when constructing the code. It must be either :code:`'high'` (move combined symbols as high as possible) or :code:`'low'` (move combined symbols as low as possible). The default value is :code:`'high'`.
        """
        self._pmf = np.array(pmf)
        self._policy = policy

        if policy not in ['high', 'low']:
            raise ValueError("Parameter 'policy' must be in {'high', 'low'}")

        super().__init__(codewords=HuffmanCode._huffman_algorithm(pmf, source_block_size, policy),
                         source_cardinality=self._pmf.size)

    @property
    def pmf(self):
        """
        The probability mass function used to construct the code. This property is read-only.
        """
        return self._pmf

    @staticmethod
    def _huffman_algorithm(pmf, source_block_size, policy):
        class Node:
            def __init__(self, index, probability):
                self.index = index
                self.probability = probability
                self.parent = None
                self.bit = None
            def __lt__(self, other):
                if policy == 'high':
                    return (self.probability, self.index) < (other.probability, other.index)
                elif policy == 'low':
                    return (self.probability, -self.index) < (other.probability, -other.index)

        tree = [Node(i, np.prod(probs)) for (i, probs) in enumerate(itertools.product(pmf, repeat=source_block_size))]
        queue = [node for node in tree]
        heapq.heapify(queue)
        while len(queue) > 1:
            node1 = heapq.heappop(queue)
            node0 = heapq.heappop(queue)
            node1.bit = 1
            node0.bit = 0
            node = Node(index=len(tree), probability=node0.probability + node1.probability)
            node0.parent = node1.parent = node.index
            heapq.heappush(queue, node)
            tree.append(node)

        codewords = []
        for symbol in range(len(pmf)**source_block_size):
            node = tree[symbol]
            bits = []
            while node.parent is not None:
                bits.insert(0, node.bit)
                node = tree[node.parent]
            codewords.append(tuple(bits))

        return codewords

    def __repr__(self):
        args = 'pmf={}, source_block_size={}'.format(self._pmf.tolist(), self._source_block_size)
        return '{}({})'.format(self.__class__.__name__, args)


class VariableToFixedCode:
    """
    Binary (prefix-free) variable-to-fixed length code. Let :math:`\\mathcal{X}` be the alphabet of some discrete source. A *binary variable-to-fixed length code* of code block size :math:`n` is defined by a (possibly partial) decoding mapping :math:`\\mathrm{Dec} : \\{ 0, 1 \\}^n \\to \\mathcal{X}^+`, where :math:`\\mathcal{X}^+` denotes the set of all finite-length, non-empty strings from the source alphabet. Here, for simplicity, the source alphabet is always taken as :math:`\\mathcal{X} = \\{0, 1, \\ldots, |\\mathcal{X} - 1| \\}`. The elements in the image of :math:`\\mathrm{Dec}` are called *sourcewords*.

    Also, we only consider *prefix-free* codes, in which no sourceword is a prefix of any other sourceword.
    """
    def __init__(self, sourcewords):
        """
        Constructor for the class. It expects the following parameters:

        :code:`sourcewords` : :obj:`list` of :obj:`tuple` of :obj:`int`
            The sourcewords of the code. Must be a list of length at most :math:`2^n` containing tuples of integers in :math:`\\mathcal{X}`. The tuple in position :math:`i` of :code:`sourcewords` should be equal to :math:`\\mathrm{Dec}(v)`, where :math:`v` is the :math:`i`-th element in the lexicographic ordering of :math:`\\{ 0, 1 \\}^n`.

        *Note:* The code block size :math:`n` is inferred from :code:`len(sourcewords)`.

        .. rubric:: Examples

        >>> code = komm.VariableToFixedCode(sourcewords=[(1,), (2,), (0,1), (0,2), (0,0,0), (0,0,1), (0,0,2)])
        >>> pprint(code.enc_mapping)
        {(0, 0, 0): (1, 0, 0),
         (0, 0, 1): (1, 0, 1),
         (0, 0, 2): (1, 1, 0),
         (0, 1): (0, 1, 0),
         (0, 2): (0, 1, 1),
         (1,): (0, 0, 0),
         (2,): (0, 0, 1)}
        >>> pprint(code.dec_mapping)
        {(0, 0, 0): (1,),
         (0, 0, 1): (2,),
         (0, 1, 0): (0, 1),
         (0, 1, 1): (0, 2),
         (1, 0, 0): (0, 0, 0),
         (1, 0, 1): (0, 0, 1),
         (1, 1, 0): (0, 0, 2)}
        """
        # TODO: Assert prefix-free
        self._sourcewords = sourcewords
        self._source_cardinality = max(itertools.chain(*sourcewords)) + 1
        self._code_block_size = (len(sourcewords) - 1).bit_length()
        self._enc_mapping = {}
        self._dec_mapping = {}
        for symbols, bits in zip(itertools.product(range(2), repeat=self._code_block_size), sourcewords):
            self._enc_mapping[bits] = tuple(symbols)
            self._dec_mapping[tuple(symbols)] = bits

    @property
    def source_cardinality(self):
        """
        The cardinality :math:`|\\mathcal{X}|` of the source alphabet.
        """
        return self._source_cardinality

    @property
    def code_block_size(self):
        """
        The code block size :math:`n`.
        """
        return self._code_block_size

    @property
    def enc_mapping(self):
        """
        The encoding mapping :math:`\\mathrm{Enc}` of the code.
        """
        return self._enc_mapping

    @property
    def dec_mapping(self):
        """
        The decoding mapping :math:`\\mathrm{Dec}` of the code.
        """
        return self._dec_mapping

    def rate(self, pmf):
        """
        Computes the expected rate :math:`R` of the code, assuming a given :term:`pmf`. It is given in bits per source symbol.

        **Input:**

        :code:`pmf` : 1D-array of :obj:`float`
            The (first-order) probability mass function :math:`p_X(x)` to be assumed.

        **Output:**

        :code:`rate` : :obj:`float`
            The expected rate :math:`R` of the code.

        .. rubric:: Examples

        >>> code = komm.VariableToFixedCode([(0,0,0), (0,0,1), (0,1), (1,)])
        >>> code.rate([2/3, 1/3])
        0.9473684210526315
        """
        probabilities = np.array([np.prod([pmf[x] for x in symbols]) for symbols in self._sourcewords])
        lengths = [len(symbols) for symbols in self._sourcewords]
        return self._code_block_size / np.dot(lengths, probabilities)

    def encode(self, symbol_sequence):
        """
        Encodes a given sequence of symbols to its corresponding sequence of bits.

        **Input:**

        :code:`symbol_sequence` : 1D-array of :obj:`int`
            The sequence of symbols to be encoded. Must be a 1D-array with elements in :math:`\\mathcal{X} = \\{0, 1, \\ldots, |\\mathcal{X} - 1| \\}`.

        **Output:**

        :code:`bit_sequence` : 1D-array of :obj:`int`
            The sequence of bits corresponding to :code:`symbol_sequence`.

        .. rubric:: Examples

        >>> code = komm.VariableToFixedCode([(0,0,0), (0,0,1), (0,1), (1,)])
        >>> code.encode([0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0])
        array([0, 0, 1, 1, 0, 1, 0, 1, 0, 1, 0, 0])
        """
        return np.array(_parse_prefix_free(symbol_sequence, self._enc_mapping))

    def decode(self, bit_sequence):
        """
        Decodes a given sequence of bits to its corresponding sequence of symbols.

        **Input:**

        :code:`bit_sequence` : 1D-array of :obj:`int`
            The sequence of bits to be decoded. Must be a 1D-array with elements in :math:`\\{ 0, 1 \\}`.  Its length must be a multiple of :math:`n`.

        **Output:**

        :code:`symbol_sequence` : 1D-array of :obj:`int`
            The sequence of symbols corresponding to :code:`bits`.

        .. rubric:: Examples

        >>> code = komm.VariableToFixedCode([(0,0,0), (0,0,1), (0,1), (1,)])
        >>> code.decode([0, 0, 1, 1, 0, 1, 0, 1, 0, 1, 0, 0])
        array([0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0])
        """
        bits_reshaped = np.reshape(bit_sequence, newshape=(-1, self._code_block_size))
        return np.concatenate([self._dec_mapping[tuple(bits)] for bits in bits_reshaped])

    def __repr__(self):
        args = 'sourcewords={}'.format(self._sourcewords)
        return '{}({})'.format(self.__class__.__name__, args)


class TunstallCode(VariableToFixedCode):
    """
    Tunstall code. It is an optimal (minimal expected rate) variable-to-fixed length code (:class:`VariableToFixedCode`) for a given probability mass function.

    .. rubric:: Examples

    >>> code = komm.TunstallCode([0.6, 0.3, 0.1], code_block_size=3)
    >>> pprint(code.enc_mapping)
    {(0, 0, 0): (0, 0, 0),
     (0, 0, 1): (0, 0, 1),
     (0, 0, 2): (0, 1, 0),
     (0, 1): (0, 1, 1),
     (0, 2): (1, 0, 0),
     (1,): (1, 0, 1),
     (2,): (1, 1, 0)}
    """
    def __init__(self, pmf, code_block_size):
        """
        Constructor for the class. It expects the following parameters:

        :code:`pmf` : 1D-array of :obj:`float`
            The probability mass function used to construct the code.

        :code:`code_block_size` : :obj:`int`, optional
            The code block size :math:`n`. Must satisfy :math:`2^n \geq |\\mathcal{X}|`, where :math:`|\\mathcal{X}|` is the cardinality of the source alphabet, given by :code:`len(pmf)`.
        """
        self._pmf = np.array(pmf)

        if 2**code_block_size < len(pmf):
            raise ValueError("Code block size is too low")

        super().__init__(sourcewords=TunstallCode._tunstall_algorithm(pmf, code_block_size))

    @property
    def pmf(self):
        """
        The probability mass function used to construct the code. This property is read-only.
        """
        return self._pmf

    @staticmethod
    def _tunstall_algorithm(pmf, code_block_size):
        class Node:
            def __init__(self, symbols, probability):
                self.symbols = symbols
                self.probability = probability
            def __lt__(self, other):
                return -self.probability < -other.probability

        queue = [Node((symbol,), probability) for (symbol, probability) in enumerate(pmf)]
        heapq.heapify(queue)

        while len(queue) + len(pmf) - 1 < 2**code_block_size:
            node = heapq.heappop(queue)
            for (symbol, probability) in enumerate(pmf):
                new_node = Node(node.symbols + (symbol,), node.probability * probability)
                heapq.heappush(queue, new_node)
        sourcewords = sorted(node.symbols for node in queue)

        return sourcewords

    def __repr__(self):
        args = 'pmf={}, code_block_size={}'.format(self._pmf.tolist(), self._code_block_size)
        return '{}({})'.format(self.__class__.__name__, args)


def _parse_prefix_free(input_sequence, dictionary):
    output_sequence = []
    i = 0
    while i < len(input_sequence):
        j = 1
        while i + j <= len(input_sequence):
            try:
                key = tuple(input_sequence[i : i + j])
                output_sequence.extend(dictionary[key])
                break
            except KeyError:
                j += 1
        i += j
    return output_sequence



import numpy as np

from ._util import \
    _entropy


__all__ = ['DiscreteMemorylessSource']


class DiscreteMemorylessSource:
    """
    Discrete memoryless source (DMS). It is defined by an *alphabet* :math:`\\mathcal{X}` and a *probability mass function* (:term:`pmf`) :math:`p_X`. Here, for simplicity, the alphabet is always taken as :math:`\\mathcal{X} = \\{ 0, 1, \\ldots, |\\mathcal{X}| - 1 \\}`. The :term:`pmf` :math:`p_X` gives the probability of the source emitting the symbol :math:`X = x`.

    To invoke the source, call the object giving the number of symbols to be emitted as parameter (see example below).
    """
    def __init__(self, pmf):
        """
        Constructor for the class. It expects the following parameter:

        :code:`pmf` : 1D-array of :obj:`float`
            The source probability mass function :math:`p_X`. The element in position :math:`x \\in \\mathcal{X}` must be equal to :math:`p_X(x)`.

        .. rubric:: Examples

        >>> dms = komm.DiscreteMemorylessSource([0.5, 0.4, 0.1])
        >>> dms(10)  #doctest:+SKIP
        array([1, 2, 1, 0, 0, 1, 1, 0, 1, 1])
        """
        self.pmf = pmf

    @property
    def pmf(self):
        """
        The source probability mass function :math:`p_X`. This is a read-and-write property.
        """
        return self._pmf

    @pmf.setter
    def pmf(self, value):
        self._pmf = np.array(value, dtype=np.float)
        self._cardinality = self._pmf.size

    @property
    def cardinality(self):
        """
        The cardinality :math:`|\\mathcal{X}|` of the source alphabet. This property is read-only.
        """
        return self._cardinality

    def entropy(self):
        """
        Returns the source entropy :math:`\\mathrm{H}(X)`.

        .. rubric:: Examples

        >>> dms = komm.DiscreteMemorylessSource([1/2, 1/4, 1/8, 1/8])
        >>> dms.entropy()
        1.75
        """
        return _entropy(self._pmf)

    def __call__(self, size):
        return np.random.choice(self._cardinality, p=self._pmf, size=size)

    def __repr__(self):
        args = 'pmf={}'.format(self._pmf.tolist())
        return '{}({})'.format(self.__class__.__name__, args)



from ._algebra import *
from ._channels import *
from ._error_control_block import *
from ._error_control_convolutional import *
from ._finite_state_machine import *
from ._modulation import *
from ._pulses import *
from ._quantization import *
from ._sequences import *
from ._source_coding import *
from ._sources import *
from ._util import *

import inspect as _inspect
import sys as _sys
for _, _cls in _inspect.getmembers(_sys.modules[__name__], _inspect.isclass):
    if hasattr(_cls, '_process_docstring'):
        _cls._process_docstring()

from pkg_resources import get_distribution as _get_distribution
__version__ = _get_distribution('komm').version



import numpy as np

from scipy import special

__all__ = ['binlist2int', 'int2binlist', 'pack', 'unpack',
           'qfunc', 'qfuncinv',  'entropy']


# Functions beggining with underscore:
# - Should not be used by the end user.
# - Should be as fast as possible.
# - May have assumptions on the input and on the output
#   (e.g., may assume the input is a list, or a numpy array, etc.).
#
# Functions without underscore:
# - Are available to the end user.
# - Should work when the input is a list, a numpy array, etc.
# - Should check the input whenever possible.
# - Should return a numpy array (intead of a list) whenever possible.

# TODO: Rename binlist2int and int2binlist to something better.
# TODO: Vectorize those functions (e.g., axis=1).


def _binlist2int(list_):
    return sum(1 << i for (i, b) in enumerate(list_) if b != 0)

def binlist2int(list_):
    """
    Converts a bit array to its integer representation.
    """
    return _binlist2int(list_)


def _int2binlist(int_, width=None):
    if width is None:
        width = max(int_.bit_length(), 1)
    return [(int_ >> i) & 1 for i in range(width)]

def int2binlist(int_, width=None):
    """
    Converts an integer to its bit array representation.
    """
    return np.array(_int2binlist(int_, width))


def _pack(list_, width):
    return np.apply_along_axis(_binlist2int, 1, np.reshape(list_, newshape=(-1, width)))

def pack(list_, width):
    """
    Packs a given integer array.
    """
    return _pack(list_, width)


def _unpack(list_, width):
    return np.ravel([_int2binlist(i, width=width) for i in list_])

def unpack(list_, width):
    """
    Unpacks a given bit array.
    """
    return _unpack(list_, width)


def _qfunc(x):
    return 0.5 * special.erfc(x / np.sqrt(2))

def qfunc(x):
    """
    Computes the gaussian Q-function. It is given by

    .. math::

       \\mathrm{Q}(x) = \\frac{1}{\\sqrt{2\\pi}} \\int_x^\\infty \\mathrm{e}^{-u^2/2} \\, \\mathrm{d}u.

    **Input:**

    :code:`x` : :obj:`float` or array of :obj:`float`
        The input to the function. May be any float or array of floats.

    **Output:**

    :code:`y` : same as input
        The value :math:`y = \\mathrm{Q}(x)`.

    .. rubric:: Examples

    >>> komm.qfunc(0.0)
    0.5

    >>> komm.qfunc([-1.0, 0.0, 1.0])
    array([0.84134475, 0.5       , 0.15865525])
    """
    return _qfunc(x)


def _qfuncinv(y):
    return np.sqrt(2) * special.erfcinv(2 * y)

def qfuncinv(x):
    """
    Computes the inverse gaussian Q-function.

    **Input:**

    :code:`y` : :obj:`float` or array of :obj:`float`
        The input to the function. Should be a float or array of floats in the real interval :math:`[0, 1]`.

    **Output:**

    :code:`x` : same as input
        The value :math:`x = \\mathrm{Q^{-1}}(y)`.

    >>> komm.qfuncinv(0.5)  #doctest:+SKIP
    0.0

    >>> komm.qfuncinv([0.841344746, 0.5, 0.158655254])  #doctest:+SKIP
    array([-1., 0.,  1.])
    """
    return _qfuncinv(np.array(x))


def _entropy_base_e(pmf):
    # Assumptions:
    # - pmf is a 1D numpy array.
    # - pmf is a valid pmf.
    return -np.dot(pmf, np.log(pmf, where=(pmf > 0)))

def _entropy_base_2(pmf):
    # Assumptions: Same as _entropy_base_e.
    return -np.dot(pmf, np.log2(pmf, where=(pmf > 0)))

def _entropy(pmf, base=2.0):
    # Assumptions: Same as _entropy_base_e.
    if base == 'e':
        return _entropy_base_e(pmf)
    elif base == 2:
        return _entropy_base_2(pmf)
    else:
        return _entropy_base_e(pmf) / np.log(base)

def entropy(pmf, base=2.0):
    """
    Computes the entropy of a random variable with a given :term:`pmf`. Let :math:`X` be a random variable with :term:`pmf` :math:`p_X` and alphabet :math:`\\mathcal{X}`. Its entropy is given by

    .. math::

       \\mathrm{H}(X) = \\sum_{x \\in \\mathcal{X}} p_X(x) \\log \\frac{1}{p_X(x)},

    By default, the base of the logarithm is :math:`2`, in which case the entropy is measured in bits.

    References: :cite:`Cover.Thomas.06` (Ch. 2)

    **Input:**

    :code:`pmf` : 1D-array of :obj:`float`
        The probability mass function :math:`p_X` of the random variable. It must be a valid :term:`pmf`, that is, all of its values must be non-negative and sum up to :math:`1`.

    :code:`base` : :obj:`float` or :obj:`str`, optional
        The base of the logarithm to be used. It must be a positive float or the string :code:`'e'`. The default value is :code:`2.0`.

    **Output:**

    :code:`entropy` : :obj:`float`
        The entropy :math:`\\mathrm{H}(X)` of the random variable.

    .. rubric:: Examples

    >>> komm.entropy([1/4, 1/4, 1/4, 1/4])
    2.0

    >>> komm.entropy([1/3, 1/3, 1/3], base=3.0)  #doctest:+SKIP
    1.0

    >>> komm.entropy([1.0, 1.0])
    Traceback (most recent call last):
     ...
    ValueError: Invalid pmf
    """
    pmf = np.array(pmf, dtype=np.float)
    if not np.allclose(np.sum(pmf), 1.0) or not np.alltrue(pmf >= 0.0):
        raise ValueError("Invalid pmf")
    return _entropy(pmf, base)


def _mutual_information(input_pmf, transition_probabilities, base=2.0):
    output_pmf = np.dot(input_pmf, transition_probabilities)
    entropy_output_prior = _entropy(output_pmf, base)
    entropy_output_posterior = np.dot(input_pmf, np.apply_along_axis(_entropy, 1, transition_probabilities))
    return entropy_output_prior - entropy_output_posterior



import itertools

import numpy as np

from ._util import int2binlist

__all__ = ['RealModulation', 'PAModulation',
           'ComplexModulation', 'ASKModulation', 'PSKModulation', 'APSKModulation', 'QAModulation']


class _Modulation:
    def __init__(self, constellation, labeling):
        self._constellation = np.array(constellation)
        self._order = self._constellation.size
        if self._order & (self._order - 1):
            raise ValueError("The length of constellation must be a power of two")
        self._bits_per_symbol = (self._order - 1).bit_length()

        self._labeling = np.array(labeling, dtype=np.int)
        if not np.array_equal(np.sort(self._labeling), np.arange(self._order)):
            raise ValueError("The labeling must be a permutation of [0 : order)")
        self._mapping = {symbol: tuple(int2binlist(label, width=self._bits_per_symbol)) \
                         for (symbol, label) in enumerate(self._labeling)}
        self._inverse_mapping = dict((value, key) for key, value in self._mapping.items())

        self._channel_snr = 1.0
        self._channel_N0 = self.energy_per_symbol / self._channel_snr

    def __repr__(self):
        args = 'constellation={}, labeling={}'.format(self._constellation.tolist(), self._labeling.tolist())
        return '{}({})'.format(self.__class__.__name__, args)

    @property
    def constellation(self):
        """
        The constellation :math:`\\mathcal{S}` of the modulation. This property is read-only.
        """
        return self._constellation

    @property
    def labeling(self):
        """
        The binary labeling :math:`\\mathcal{Q}` of the modulation. This property is read-only.
        """
        return self._labeling

    @property
    def order(self):
        """
        The order :math:`M` of the modulation. This property is read-only.
        """
        return self._order

    @property
    def bits_per_symbol(self):
        """
        The number :math:`m` of bits per symbol of the modulation. It is given by :math:`m = \\log_2 M`, where :math:`M` is the order of the modulation. This property is read-only.
        """
        return self._bits_per_symbol

    @property
    def energy_per_symbol(self):
        """
        The average symbol energy :math:`E_\\mathrm{s}` of the constellation. It assumes equiprobable symbols. It is given by

        .. math::

            E_\\mathrm{s} = \\frac{1}{M} \\sum_{s_i \\in \\mathcal{S}} |s_i|^2,

        where :math:`|s_i|^2` is the energy of symbol :math:`s_i \\in \\mathcal{S}` and :math:`M` is the order of the modulation. This property is read-only.
        """
        return np.real(np.dot(self._constellation, self._constellation.conj())) / self._order

    @property
    def energy_per_bit(self):
        """
        The average bit energy :math:`E_\\mathrm{b}` of the constellation. It assumes equiprobable symbols. It is given by :math:`E_\\mathrm{b} = E_\\mathrm{s} / m`, where :math:`E_\\mathrm{s}` is the average symbol energy, and :math:`m` is the number of bits per symbol of the modulation.
        """
        return self.energy_per_symbol / np.log2(self._order)

    @property
    def minimum_distance(self):
        """
        The minimum euclidean distance of the constellation.
        """
        pass

    @property
    def channel_snr(self):
        """
        The signal-to-noise ratio :math:`\\mathrm{SNR}` of the channel. This is used in soft-decision methods. This is a read-and-write property.
        """
        return self._channel_snr

    @channel_snr.setter
    def channel_snr(self, value):
        self._channel_snr = value
        self._channel_N0 = self.energy_per_symbol / self._channel_snr

    def bits_to_symbols(self, bits):
        """
        Converts bits to symbols using the modulation binary labeling.

        **Input:**

        :code:`bits` : 1D-array of :obj:`int`
            The bits to be converted. It should be a 1D-array of integers in the set :math:`\\{ 0, 1 \\}`. Its length must be a multiple of :math:`m`.

        **Output:**

        :code:`symbols` : 1D-array of :obj:`int`
            The symbols corresponding to :code:`bits`. It is a 1D-array of integers in the set :math:`[0 : M)`. Its length is equal to the length of :code:`bits` divided by :math:`m`.
        """
        m = self._bits_per_symbol
        n_symbols = len(bits) // m
        assert len(bits) == n_symbols * m
        symbols = np.empty(n_symbols, dtype=np.int)
        for i, bit_sequence in enumerate(np.reshape(bits, newshape=(n_symbols, m))):
            symbols[i] = self._inverse_mapping[tuple(bit_sequence)]
        return symbols

    def symbols_to_bits(self, symbols):
        """
        Converts symbols to bits using the modulation binary labeling.

        **Input:**

        :code:`symbols` : 1D-array of :obj:`int`
            The symbols to be converted. It should be a 1D-array of integers in the set :math:`[0 : M)`. It may be of any length.

        **Output:**

        :code:`bits` : 1D-array of :obj:`int`
            The bits corresponding to :code:`symbols`. It is a 1D-array of integers in the set :math:`\\{ 0, 1 \\}`. Its length is equal to the length of :code:`symbols` multiplied by :math:`m`.
        """
        m = self._bits_per_symbol
        n_bits = len(symbols) * m
        bits = np.empty(n_bits, dtype=np.int)
        for i, symbol in enumerate(symbols):
            bits[i*m : (i + 1)*m] = self._mapping[symbol]
        return bits

    def modulate(self, bits):
        """
        Modulates a sequence of bits to its corresponding constellation points.
        """
        symbols = self.bits_to_symbols(bits)
        return self._constellation[symbols]

    def _hard_symbol_demodulator(self, received):
        """General minimum distance hard demodulator."""
        mpoints, mconst = np.meshgrid(received, self._constellation)
        return np.argmin(np.absolute(mpoints - mconst), axis=0)

    def _soft_bit_demodulator(self, received):
        """Computes L-values of received points"""
        m = self._bits_per_symbol

        def pdf_received_given_bit(bit_index, bit_value):
            bits = np.empty(m, dtype=np.int)
            bits[bit_index] = bit_value
            rest_index = np.setdiff1d(np.arange(m), [bit_index])
            f = 0.0
            for b_rest in itertools.product([0, 1], repeat=m-1):
                bits[rest_index] = b_rest
                point = self._constellation[self._inverse_mapping[tuple(bits)]]
                f += np.exp(-np.abs(received - point)**2 / self._channel_N0)
            return f

        soft_bits = np.empty(len(received)*m, dtype=np.float)
        for bit_index in range(m):
            p0 = pdf_received_given_bit(bit_index, 0)
            p1 = pdf_received_given_bit(bit_index, 1)
            soft_bits[bit_index::m] = np.log(p0 / p1)

        return soft_bits

    def demodulate(self, received, decision_method='hard'):
        """
        Demodulates a sequence of received points to a sequence of bits.
        """
        if decision_method == 'hard':
            symbols_hat = self._hard_symbol_demodulator(received)
            return self.symbols_to_bits(symbols_hat)
        elif decision_method == 'soft':
            return self._soft_bit_demodulator(received)
        else:
            raise ValueError("Parameter 'decision_method' should be either 'hard' or 'soft'")

    @staticmethod
    def _labeling_natural(order):
        labeling = np.arange(order)
        return labeling

    @staticmethod
    def _labeling_reflected(order):
        labeling = np.arange(order)
        labeling ^= (labeling >> 1)
        return labeling

    @staticmethod
    def _labeling_reflected_2d(order_I, order_Q):
        labeling_I = _Modulation._labeling_reflected(order_I)
        labeling_Q = _Modulation._labeling_reflected(order_Q)
        labeling = np.empty(order_I * order_Q, dtype=np.int)
        for i, (i_Q, i_I) in enumerate(itertools.product(labeling_Q, labeling_I)):
            labeling[i] = i_I + order_I * i_Q
        return labeling


class RealModulation(_Modulation):
    """
    General real modulation scheme. A *real modulation scheme* of order :math:`M` is defined by a *constellation* :math:`\\mathcal{S}`, which is an ordered subset (a list) of real numbers, with :math:`|\\mathcal{S}| = M`, and a *binary labeling* :math:`\\mathcal{Q}`, which is a permutation of :math:`[0: M)`. The order :math:`M` of the modulation must be a power of :math:`2`.
    """
    def __init__(self, constellation, labeling):
        """
        Constructor for the class. It expects the following parameters:

        :code:`constellation` : 1D-array of :obj:`complex`
            The constellation :math:`\\mathcal{S}` of the modulation. Must be a 1D-array containing :math:`M` real numbers.

        :code:`labeling` : 1D-array of :obj:`int`
            The binary labeling :math:`\\mathcal{Q}` of the modulation. Must be a 1D-array of integers corresponding to a permutation of :math:`[0 : M)`.

        .. rubric:: Examples

        >>> mod = komm.RealModulation(constellation=[-0.5, 0, 0.5, 2], labeling=[0, 1, 3, 2])
        >>> mod.constellation
        array([-0.5,  0. ,  0.5,  2. ])
        >>> mod.modulate([0, 0, 1, 1, 0, 0, 1, 0, 1, 0])
        array([-0.5,  0.5, -0.5,  0. ,  0. ])
        """
        super().__init__(np.array(constellation, dtype=np.float), labeling)


class PAModulation(RealModulation):
    """
    Pulse-amplitude modulation (PAM). It is a real modulation scheme (:class:`RealModulation`) in which the points of the constellation :math:`\\mathcal{S}` are *uniformly arranged* in the real line. More precisely,

    .. math::
        \\mathcal{S} = \\{ \\pm (2i + 1)A : i \\in [0 : M) \\},

    where :math:`M` is the *order* (a power of :math:`2`), and :math:`A` is the *base amplitude*. The PAM constellation is depicted below for :math:`M = 8`.

    |

    .. image:: figures/pam_8.png
       :alt: 8-PAM constellation.
       :align: center
    """
    def __init__(self, order, base_amplitude=1.0, labeling='reflected'):
        """
        Constructor for the class. It expects the following parameters:

        :code:`order` : :obj:`int`
            The order :math:`M` of the modulation. It must be a power of :math:`2`.

        :code:`base_amplitude` : :obj:`float`, optional
            The base amplitude :math:`A` of the constellation. The default value is :code:`1.0`.

        :code:`labeling` : (1D-array of :obj:`int`) or :obj:`str`, optional
            The binary labeling :math:`\\mathcal{Q}` of the modulation. Can be specified either as a 1D-array of integers, in which case must be permutation of :math:`[0 : M)`, or as a string, in which case must be one of :code:`'natural'` or :code:`'reflected'`. The default value is :code:`'reflected'` (Gray code).

        .. rubric:: Examples

        >>> pam = komm.PAModulation(4, base_amplitude=2.0)
        >>> pam.constellation
        array([-6., -2.,  2.,  6.])
        >>> pam.labeling
        array([0, 1, 3, 2])
        >>> pam.modulate([0, 0, 1, 1, 0, 0, 1, 0, 1, 0])
        array([-6.,  2., -6., -2., -2.])

        """
        constellation = base_amplitude * np.arange(-order + 1, order, step=2, dtype=np.int)

        if isinstance(labeling, str):
            if labeling in ['natural', 'reflected']:
                labeling = getattr(_Modulation, '_labeling_' + labeling)(order)
            else:
                raise ValueError("Only 'natural' or 'reflected' are supported for {}".format(self.__class__.__name__))

        super().__init__(constellation, labeling)

        self._base_amplitude = float(base_amplitude)

    def __repr__(self):
        args = '{}, base_amplitude={}'.format(self._order, self._base_amplitude)
        return '{}({})'.format(self.__class__.__name__, args)


class ComplexModulation(_Modulation):
    """
    General complex modulation scheme. A *complex modulation scheme* of order :math:`M` is defined by a *constellation* :math:`\\mathcal{S}`, which is an ordered subset (a list) of complex numbers, with :math:`|\\mathcal{S}| = M`, and a *binary labeling* :math:`\\mathcal{Q}`, which is a permutation of :math:`[0: M)`. The order :math:`M` of the modulation must be a power of :math:`2`.
    """
    def __init__(self, constellation, labeling):
        """
        Constructor for the class. It expects the following parameters:

        :code:`constellation` : 1D-array of :obj:`complex`
            The constellation :math:`\\mathcal{S}` of the modulation. Must be a 1D-array containing :math:`M` complex numbers.

        :code:`labeling` : 1D-array of :obj:`int`
            The binary labeling :math:`\\mathcal{Q}` of the modulation. Must be a 1D-array of integers corresponding to a permutation of :math:`[0 : M)`.

        .. rubric:: Examples

        >>> mod = komm.ComplexModulation(constellation=[0.0, -1, 1, 1j], labeling=[0, 1, 2, 3])
        >>> mod.constellation
        array([ 0.+0.j, -1.+0.j,  1.+0.j,  0.+1.j])
        >>> mod.modulate([0, 0, 1, 1, 0, 0, 1, 0, 1, 0])
        array([ 0.+0.j,  0.+1.j,  0.+0.j, -1.+0.j, -1.+0.j])
        """
        super().__init__(np.array(constellation, dtype=np.complex), labeling)


class ASKModulation(ComplexModulation):
    """
    Amplitude-shift keying (ASK) modulation. It is a complex modulation scheme (:class:`ComplexModulation`) in which the points of the constellation :math:`\\mathcal{S}` are *uniformly arranged* in a ray. More precisely,

    .. math::

        \\mathcal{S} = \\{ iA \\exp(\\mathrm{j}\\phi): i \\in [0 : M) \\},

    where :math:`M` is the *order* (a power of :math:`2`), :math:`A` is the *base amplitude*, and :math:`\\phi` is the *phase offset* of the modulation.  The ASK constellation is depicted below for :math:`M = 4`.

    .. image:: figures/ask_4.png
       :alt: 4-ASK constellation.
       :align: center
    """
    def __init__(self, order, base_amplitude=1.0, phase_offset=0.0, labeling='reflected'):
        """
        Constructor for the class. It expects the following parameters:

        :code:`order` : :obj:`int`
            The order :math:`M` of the modulation. It must be a power of :math:`2`.

        :code:`base_amplitude` : :obj:`float`, optional
            The base amplitude :math:`A` of the constellation. The default value is :code:`1.0`.

        :code:`phase_offset` : :obj:`float`, optional
            The phase offset :math:`\\phi` of the constellation. The default value is :code:`0.0`.

        :code:`labeling` : (1D-array of :obj:`int`) or :obj:`str`, optional
            The binary labeling :math:`\\mathcal{Q}` of the modulation. Can be specified either as a 1D-array of integers, in which case must be permutation of :math:`[0 : M)`, or as a string, in which case must be one of :code:`'natural'` or :code:`'reflected'`. The default value is :code:`'reflected'` (Gray code).

        .. rubric:: Examples

        >>> ask = komm.ASKModulation(4, base_amplitude=2.0)
        >>> ask.constellation
        array([0.+0.j, 2.+0.j, 4.+0.j, 6.+0.j])
        >>> ask.labeling
        array([0, 1, 3, 2])
        >>> ask.modulate([0, 0, 1, 1, 0, 0, 1, 0, 1, 0])
        array([0.+0.j, 4.+0.j, 0.+0.j, 2.+0.j, 2.+0.j])
        >>> ask.demodulate([(0.99+0.3j), (1.01-0.5j), (4.99+0.7j), (5.01-0.9j)])
        array([0, 0, 1, 0, 1, 1, 0, 1])
        """
        constellation = base_amplitude * np.arange(order, dtype=np.int) * np.exp(1j*phase_offset)

        if isinstance(labeling, str):
            if labeling in ['natural', 'reflected']:
                labeling = getattr(_Modulation, '_labeling_' + labeling)(order)
            else:
                raise ValueError("Only 'natural' or 'reflected' are supported for {}".format(self.__class__.__name__))

        super().__init__(constellation, labeling)

        self._base_amplitude = float(base_amplitude)
        self._phase_offset = float(phase_offset)

    def __repr__(self):
        args = '{}, base_amplitude={}, phase_offset={}'.format(self._order, self._base_amplitude, self._phase_offset)
        return '{}({})'.format(self.__class__.__name__, args)


class PSKModulation(ComplexModulation):
    """
    Phase-shift keying (PSK) modulation. It is a complex modulation scheme (:class:`ComplexModulation`) in which the points of the constellation :math:`\\mathcal{S}` are *uniformly arranged* in a circle. More precisely,

    .. math::
        \\mathcal{S} = \\left \\{ A \\exp \\left( \\mathrm{j} \\frac{2 \\pi i}{M} \\right) \\exp(\\mathrm{j} \\phi) : i \\in [0 : M) \\right \\}

    where :math:`M` is the *order* (a power of :math:`2`), :math:`A` is the *amplitude*, and :math:`\\phi` is the *phase offset* of the modulation. The PSK constellation is depicted below for :math:`M = 8`.

    .. image:: figures/psk_8.png
       :alt: 8-PSK constellation.
       :align: center
    """
    def __init__(self, order, amplitude=1.0, phase_offset=0.0, labeling='reflected'):
        """
        Constructor for the class. It expects the following parameters:

        :code:`order` : :obj:`int`
            The order :math:`M` of the modulation. It must be a power of :math:`2`.

        :code:`amplitude` : :obj:`float`, optional
            The amplitude :math:`A` of the constellation. The default value is :code:`1.0`.

        :code:`phase_offset` : :obj:`float`, optional
            The phase offset :math:`\\phi` of the constellation. The default value is :code:`0.0`.

        :code:`labeling` : (1D-array of :obj:`int`) or :obj:`str`, optional
            The binary labeling :math:`\\mathcal{Q}` of the modulation. Can be specified either as a 1D-array of integers, in which case must be permutation of :math:`[0 : M)`, or as a string, in which case must be one of :code:`'natural'` or :code:`'reflected'`. The default value is :code:`'reflected'` (Gray code).

        .. rubric:: Examples

        >>> psk = komm.PSKModulation(4, phase_offset=np.pi/4)
        >>> psk.constellation  #doctest: +NORMALIZE_WHITESPACE
        array([ 0.70710678+0.70710678j, -0.70710678+0.70710678j, -0.70710678-0.70710678j,  0.70710678-0.70710678j])
        >>> psk.labeling
        array([0, 1, 3, 2])
        >>> psk.modulate([0, 0, 1, 1, 0, 0, 1, 0, 1, 0])  #doctest: +NORMALIZE_WHITESPACE
        array([ 0.70710678+0.70710678j, -0.70710678-0.70710678j,  0.70710678+0.70710678j, -0.70710678+0.70710678j, -0.70710678+0.70710678j])
        """
        constellation = amplitude * np.exp(2j*np.pi*np.arange(order) / order) * np.exp(1j * phase_offset)

        if isinstance(labeling, str):
            if labeling in ['natural', 'reflected']:
                labeling = getattr(_Modulation, '_labeling_' + labeling)(order)
            else:
                raise ValueError("Only 'natural' or 'reflected' are supported for {}".format(self.__class__.__name__))

        super().__init__(constellation, labeling)

        self._amplitude = float(amplitude)
        self._phase_offset = float(phase_offset)

    def __repr__(self):
        args = '{}, amplitude={}, phase_offset={}'.format(self._order, self._amplitude, self._phase_offset)
        return '{}({})'.format(self.__class__.__name__, args)


class APSKModulation(ComplexModulation):
    """
    Amplitude- and phase-shift keying (APSK) modulation. It is a complex modulation scheme (:class:`ComplexModulation`) in which the constellation is the union of component PSK constellations (:class:`PSKModulation`), called *rings*. More precisely,

    .. math::
        \\mathcal{S} = \\bigcup_{k \\in [0 : K)} \\mathcal{S}_k,

    where :math:`K` is the number of rings and

    .. math::

        \\mathcal{S}_k = \\left \\{ A_k \\exp \\left( \\mathrm{j} \\frac{2 \\pi i}{M_k} \\right) \\exp(\\mathrm{j} \\phi_k) : i \\in [0 : M_k) \\right \\},

    where :math:`M_k` is the *order*, :math:`A_k` is the *amplitude*, and :math:`\\phi_k` is the *phase offset* of the :math:`k`-th ring, for :math:`k \\in [0 : K)`. The size of the resulting complex-valued constellation is :math:`M = M_0 + M_1 + \\cdots + M_{K-1}`. The order :math:`M_k` of each ring need not be a power of :math:`2`; however, the order :math:`M` of the constructed APSK modulation must be. The APSK constellation is depicted below for :math:`(M_0, M_1) = (8, 8)` with :math:`(A_0, A_1) = (A, 2A)` and :math:`(\\phi_0, \\phi_1) = (0, \\pi/8)`.

    .. image:: figures/apsk_16.png
       :alt: 16-APSK constellation.
       :align: center
    """
    def __init__(self, orders, amplitudes, phase_offsets=0.0, labeling='natural'):
        """
        Constructor for the class. It expects the following parameters:

        :code:`orders` : :obj:`tuple` of :obj:`int`
            A :math:`K`-tuple with the orders :math:`M_k` of each ring, for :math:`k \\in [0 : K)`. The sum :math:`M_0 + M_1 + \\cdots + M_{K-1}` must be a power of :math:`2`.

        :code:`amplitudes` : :obj:`tuple` of :obj:`float`
            A :math:`K`-tuple with the amplitudes :math:`A_k` of each ring, for :math:`k \\in [0 : K)`.

        :code:`phase_offsets` : (:obj:`tuple` of :obj:`float`) or :obj:`float`, optional
            A :math:`K`-tuple with the phase offsets :math:`\\phi_k` of each ring, for :math:`k \\in [0 : K)`. If specified as a single float :math:`\\phi`, then it is assumed that :math:`\\phi_k = \\phi` for all :math:`k \\in [0 : K)`. The default value is :code:`0.0`.

        :code:`labeling` : (1D-array of :obj:`int`) or :obj:`str`, optional
            The binary labeling :math:`\\mathcal{Q}` of the modulation. Can be specified either as a 1D-array of integers, in which case must be permutation of :math:`[0 : M)`, or as a string, in which case must be equal to :code:`'natural'`. The default value is :code:`'natural'`.

        .. rubric:: Examples

        >>> apsk = komm.APSKModulation(orders=(8,8), amplitudes=(1.0, 2.0), phase_offsets=(0.0, np.pi/8))
        >>> np.round(apsk.constellation, 4)  #doctest: +SKIP
        array([ 1.    +0.j    ,  0.7071+0.7071j,  0.    +1.j    , -0.7071+0.7071j,
               -1.    +0.j    , -0.7071-0.7071j, -0.    -1.j    ,  0.7071-0.7071j,
                1.8478+0.7654j,  0.7654+1.8478j, -0.7654+1.8478j, -1.8478+0.7654j,
               -1.8478-0.7654j, -0.7654-1.8478j,  0.7654-1.8478j,  1.8478-0.7654j])
        """
        if isinstance(phase_offsets, (tuple, list)):
            phase_offsets = tuple(float(phi_k) for phi_k in phase_offsets)
            self._phase_offsets = phase_offsets
        else:
            self._phase_offsets = float(phase_offsets)
            phase_offsets = (float(phase_offsets), ) * len(orders)

        constellation = []
        for (M_k, A_k, phi_k) in zip(orders, amplitudes, phase_offsets):
            ring_constellation = A_k * np.exp(2j*np.pi*np.arange(M_k) / M_k) * np.exp(1j * phi_k)
            constellation = np.append(constellation, ring_constellation)

        order = int(np.sum(orders))
        if isinstance(labeling, str):
            if labeling in ['natural', 'reflected']:
                labeling = getattr(_Modulation, '_labeling_' + labeling)(order)
            else:
                raise ValueError("Only 'natural' or 'reflected' are supported for {}".format(self.__class__.__name__))

        super().__init__(constellation, labeling)

        self._orders = tuple(int(M_k) for M_k in orders)
        self._amplitudes = tuple(float(A_k) for A_k in amplitudes)

    def __repr__(self):
        args = '{}, amplitudes={}, phase_offsets={}'.format(self._orders, self._amplitudes, self._phase_offsets)
        return '{}({})'.format(self.__class__.__name__, args)



class QAModulation(ComplexModulation):
    """
    Quadratude-amplitude modulation (QAM). It is a complex modulation scheme (:class:`ComplexModulation`) in which the constellation :math:`\\mathcal{S}` is given as a cartesian product of two PAM (:class:`PAModulation`) constellations, namely, the *in-phase constellation*, and the *quadrature constellation*. More precisely,

    .. math::
        \\mathcal{S} = \\{ [\\pm(2i_\\mathrm{I} + 1)A_\\mathrm{I} \\pm \\mathrm{j}(2i_\\mathrm{Q} + 1)A_\\mathrm{Q}] \\exp(\\mathrm{j}\\phi) : i_\\mathrm{I} \\in [0 : M_\\mathrm{I}), i_\\mathrm{Q} \\in [0 : M_\\mathrm{Q}) \\},

    where :math:`M_\\mathrm{I}` and :math:`M_\\mathrm{Q}` are the *orders* (powers of :math:`2`), and :math:`A_\\mathrm{I}` and :math:`A_\\mathrm{Q}` are the *base amplitudes* of the in-phase and quadratude constellations, respectively. Also, :math:`\\phi` is the *phase offset*. The size of the resulting complex-valued constellation is :math:`M = M_\\mathrm{I} M_\\mathrm{Q}`, a power of :math:`2`. The QAM constellation is depicted below for :math:`(M_\\mathrm{I}, M_\\mathrm{Q}) = (4, 4)` with :math:`A_\\mathrm{I} = A_\\mathrm{Q} = A`, and for :math:`(M_\\mathrm{I}, M_\\mathrm{Q}) = (4, 2)` with :math:`A_\\mathrm{I} = A` and :math:`A_\\mathrm{Q} = 2A`.

    .. rst-class:: centered

       |fig1| |quad| |quad| |quad| |quad| |fig2|

    .. |fig1| image:: figures/qam_16.png
       :alt: 16-QAM constellation

    .. |fig2| image:: figures/qam_8.png
       :alt: 8-QAM constellation

    .. |quad| unicode:: 0x2001
       :trim:
    """
    def __init__(self, orders, base_amplitudes=1.0, phase_offset=0.0, labeling='reflected_2d'):
        """
        Constructor for the class. It expects the following parameters:

        :code:`orders` : :obj:`(int, int)` or :obj:`int`
            A tuple :math:`(M_\\mathrm{I}, M_\\mathrm{Q})` with the orders of the in-phase and quadrature constellations, respectively; both :math:`M_\\mathrm{I}` and :math:`M_\\mathrm{Q}` must be powers of :math:`2`. If specified as a single integer :math:`M`, then it is assumed that :math:`M_\\mathrm{I} = M_\\mathrm{Q} = \\sqrt{M}`; in this case, :math:`M` must be an square power of :math:`2`.

        :code:`base_amplitudes` : :obj:`(float, float)` or :obj:`float`, optional
            A tuple :math:`(A_\\mathrm{I}, A_\\mathrm{Q})` with the base amplitudes of the in-phase and quadrature constellations, respectively.  If specified as a single float :math:`A`, then it is assumed that :math:`A_\\mathrm{I} = A_\\mathrm{Q} = A`. The default value is :math:`1.0`.

        :code:`phase_offset` : :obj:`float`, optional
            The phase offset :math:`\\phi` of the constellation. The default value is :code:`0.0`.

        :code:`labeling` : (1D-array of :obj:`int`) or :obj:`str`, optional
            The binary labeling :math:`\\mathcal{Q}` of the modulation. Can be specified either as a 1D-array of integers, in which case must be permutation of :math:`[0 : M)`, or as a string, in which case must be one of :code:`'natural'` or :code:`'reflected_2d'`. The default value is :code:`'reflected_2d'` (Gray code).

        .. rubric:: Examples

        >>> qam = komm.QAModulation(16)
        >>> qam.constellation  #doctest: +NORMALIZE_WHITESPACE
        array([-3.-3.j, -1.-3.j,  1.-3.j,  3.-3.j,
               -3.-1.j, -1.-1.j,  1.-1.j,  3.-1.j,
               -3.+1.j, -1.+1.j,  1.+1.j,  3.+1.j,
               -3.+3.j, -1.+3.j,  1.+3.j,  3.+3.j])
        >>> qam.labeling
        array([ 0,  1,  3,  2,  4,  5,  7,  6, 12, 13, 15, 14,  8,  9, 11, 10])
        >>> qam.modulate([0, 0, 1, 1, 0, 0, 1, 0])
        array([-3.+1.j, -3.-1.j])

        >>> qam = komm.QAModulation(orders=(4, 2), base_amplitudes=(1.0, 2.0))
        >>> qam.constellation
        array([-3.-2.j, -1.-2.j,  1.-2.j,  3.-2.j, -3.+2.j, -1.+2.j,  1.+2.j,
                3.+2.j])
        >>> qam.labeling
        array([0, 1, 3, 2, 4, 5, 7, 6])
        >>> qam.modulate([0, 0, 1, 1, 0, 0, 1, 0, 1])
        array([-3.+2.j, -1.-2.j, -1.+2.j])
        """
        if isinstance(orders, (tuple, list)):
            order_I, order_Q = int(orders[0]), int(orders[1])
            self._orders = (order_I, order_Q)
        else:
            order_I = order_Q = int(np.sqrt(orders))
            self._orders = int(orders)

        if isinstance(base_amplitudes, (tuple, list)):
            base_amplitude_I, base_amplitude_Q = float(base_amplitudes[0]), float(base_amplitudes[1])
            self._base_amplitudes = (base_amplitude_I, base_amplitude_Q)
        else:
            base_amplitude_I = base_amplitude_Q = float(base_amplitudes)
            self._base_amplitudes = base_amplitude_I

        constellation_I = base_amplitude_I * np.arange(-order_I + 1, order_I, step=2, dtype=np.int)
        constellation_Q = base_amplitude_Q * np.arange(-order_Q + 1, order_Q, step=2, dtype=np.int)
        constellation = (constellation_I + 1j*constellation_Q[np.newaxis].T).flatten() * np.exp(1j * phase_offset)

        if isinstance(labeling, str):
            if labeling == 'natural':
                labeling = _Modulation._labeling_natural(order_I * order_Q)
            elif labeling == 'reflected_2d':
                labeling = _Modulation._labeling_reflected_2d(order_I, order_Q)
            else:
                raise ValueError("Only 'natural' or 'reflected_2d' are supported for {}".format(self.__class__.__name__))

        super().__init__(constellation, labeling)

        self._orders = orders
        self._base_amplitudes = base_amplitudes
        self._phase_offset = float(phase_offset)

    def __repr__(self):
        args = '{}, base_amplitudes={}, phase_offset={}'.format(self._orders, self._base_amplitudes, self._phase_offset)
        return '{}({})'.format(self.__class__.__name__, args)



def uniform_real_hard_demodulator(received, order):
    return np.clip(np.around((received + order - 1) / 2), 0, order - 1).astype(np.int)


def uniform_real_soft_bit_demodulator(received, snr):
    return -4 * snr * received


def ask_hard_demodulator(received, order):
    return np.clip(np.around((received.real + order - 1) / 2), 0, order - 1).astype(np.int)


def psk_hard_demodulator(received, order):
    phase_in_turns = np.angle(received) / (2 * np.pi)
    return np.mod(np.around(phase_in_turns * order).astype(np.int), order)


def bpsk_soft_bit_demodulator(received, snr):
    return 4 * snr * received.real


def qpsk_soft_bit_demodulator_reflected(received, snr):
    received_rotated = received * np.exp(2j * np.pi / 8)
    soft_bits = np.empty(2*received.size, dtype=np.float)
    soft_bits[0::2] = np.sqrt(8) * snr * received_rotated.real
    soft_bits[1::2] = np.sqrt(8) * snr * received_rotated.imag
    return soft_bits


def rectangular_hard_demodulator(received, order):
    L = int(np.sqrt(order))
    s_real = np.clip(np.around((received.real + L - 1) / 2), 0, L - 1).astype(np.int)
    s_imag = np.clip(np.around((received.imag + L - 1) / 2), 0, L - 1).astype(np.int)
    return s_real + L * s_imag



import functools
import operator

from fractions import Fraction

import numpy as np

from ._util import \
    _int2binlist, _binlist2int

__all__ = ['BinaryPolynomial', 'BinaryPolynomialFraction', 'FiniteBifield',
           'RationalPolynomial', 'RationalPolynomialFraction']


class BinaryPolynomial:
    """
    Binary polynomial. A *binary polynomial* is a polynomial whose coefficients are elements in the finite field :math:`\\mathbb{F}_2 = \\{ 0, 1 \\}`. The default constructor takes an :obj:`int` as input, whose binary digits represent the coefficients of the polynomial---the leftmost bit standing for the highest degree term. For example, the binary polynomial :math:`X^4 + X^3 + X` is represented by the integer :code:`0b11010` = :code:`0o32` = :code:`26`. There are two alternative constructors for this class, the class methods :func:`from_coefficients` and :func:`from_exponents`.  See their documentation for details.

    This class supports addition, multiplication, and division of binary polynomials.

    .. rubric:: Examples

    >>> poly = komm.BinaryPolynomial(0b11010)  # X^4 + X^3 + X
    >>> poly
    BinaryPolynomial(0b11010)

    >>> poly1 = komm.BinaryPolynomial(0b10100)  # X^4 + X^2
    >>> poly2 = komm.BinaryPolynomial(0b11010)  # X^4 + X^3 + X
    >>> poly1 + poly2  # X^3 + X^2 + X
    BinaryPolynomial(0b1110)
    >>> poly1 * poly2  # X^8 + X^7 + X^6 + X^3
    BinaryPolynomial(0b111001000)
    >>> poly1**2  # X^8 + X^4
    BinaryPolynomial(0b100010000)
    """
    def __init__(self, integer):
        self._integer = int(integer)

    @classmethod
    def from_coefficients(cls, coefficients):
        """
        Constructs a :obj:`BinaryPolynomial` from its coefficients. It expects the following parameter:

        :code:`coefficients` : 1D-array of :obj:`int`
            The coefficients of the binary polynomial---the :math:`i`-th element of the array standing for the coefficient of :math:`X^i`. For example, :code:`[0, 1, 0, 1, 1]` represents the binary polynomial :math:`X^4 + X^3 + X`.

        .. rubric:: Examples

        >>> komm.BinaryPolynomial.from_coefficients([0, 1, 0, 1, 1])  # X^4 + X^3 + X
        BinaryPolynomial(0b11010)
        """
        return cls(_binlist2int(coefficients))

    @classmethod
    def from_exponents(cls, exponents):
        """
        Constructs a :obj:`BinaryPolynomial` from its exponents. It expects the following parameter:

        :code:`coefficients` : 1D-array of :obj:`int`
            The exponents of the nonzero terms of the binary polynomial. For example, :code:`[1, 3, 4]` represents the binary polynomial :math:`X^4 + X^3 + X`.

        .. rubric:: Examples

        >>> komm.BinaryPolynomial.from_exponents([1, 3, 4])  # X^4 + X^3 + X
        BinaryPolynomial(0b11010)
        """
        return cls(_binlist2int(np.bincount(exponents)))

    @property
    def degree(self):
        """
        The degree of the polynomial. This property is read-only.

        .. rubric:: Examples

        >>> poly = komm.BinaryPolynomial(0b11010)  # X^4 + X^3 + X
        >>> poly.degree
        4
        """
        return self._integer.bit_length() - 1

    def coefficients(self, width=None):
        """
        Returns the coefficients of the binary polynomial.

        **Input:**

        :code:`width` : :obj:`int`, optional
            If this parameter is specified, the output will be filled with zeros on the right so that the its length will be the specified value.

        **Output:**

        :code:`coefficients` : 1D-array of :obj:`int`
            Coefficients of the binary polynomial. The :math:`i`-th element of the array stands for the coefficient of :math:`X^i`.

        .. rubric:: Examples

        >>> poly = komm.BinaryPolynomial(0b11010)  # X^4 + X^3 + X
        >>> poly.coefficients()
        array([0, 1, 0, 1, 1])
        >>> poly.coefficients(width=8)
        array([0, 1, 0, 1, 1, 0, 0, 0])
        """
        return np.array(_int2binlist(self._integer, width=width), dtype=np.int)

    def exponents(self):
        """
        Returns the exponents of the binary polynomial.

        **Output:**

        :code:`exponents` : 1D-array of :obj:`int`
            Exponents of the nonzero terms of the binary polynomial. The exponents are returned in ascending order.

        .. rubric:: Examples

        >>> poly = komm.BinaryPolynomial(0b11010)  # X^4 + X^3 + X
        >>> poly.exponents()
        array([1, 3, 4])
        """
        return np.flatnonzero(self.coefficients())

    def __int__(self):
        return self._integer

    def __hash__(self):
        return self._integer

    def __eq__(self, other):
        return int(self) == int(other)

    def __lshift__(self, n):
        return self.__class__(self._integer.__lshift__(n))

    def __rshift__(self, n):
        return self.__class__(self._integer.__rshift__(n))

    def __add__(self, other):
        return self.__class__(self._integer.__xor__(other._integer))

    def __sub__(self, other):
        return self.__class__(self._integer.__xor__(other._integer))

    def __mul__(self, other):
        return self.from_coefficients(np.convolve(self.coefficients(), other.coefficients()) % 2)

    def __pow__(self, exponent):
        return power(self, exponent, self.__class__)

    def __divmod__(self, den):
        div, mod, den = 0, self._integer, den._integer
        d = mod.bit_length() - den.bit_length()
        while d >= 0:
            div ^= (1 << d)
            mod ^= (den << d)
            d = mod.bit_length() - den.bit_length()
        return self.__class__(div), self.__class__(mod)

    def __floordiv__(self, other):
        return self.__divmod__(other)[0]

    def __mod__(self, other):
        return self.__divmod__(other)[1]

    def evaluate(self, point):
        """
        Evaluates the polynomial at a given point. Uses Horner's method.

        **Input:**

        :code:`point` : ring-like type
            Any Python object supporting the operations of addition, subtraction, and multiplication.

        **Output:**

        :code:`result` : ring-like type
            The result of evaluating the binary polynomial at :code:`point`. It has the same type as :code:`point`.

        .. rubric:: Examples

        >>> poly = komm.BinaryPolynomial(0b11010)  # X^4 + X^3 + X
        >>> poly.evaluate(7)  # same as 7**4 + 7**3 + 7
        2751
        >>> point = np.array([[1, 2], [3, 4]])
        >>> poly.evaluate(point)  # same as point**4 + point**3 + point
        array([[  3,  26],
               [111, 324]])
        """
        return binary_horner(self, point)

    def __repr__(self):
        args = '{}'.format(bin(self._integer))
        return '{}({})'.format(self.__class__.__name__, args)

    def __str__(self):
        return bin(self._integer)

    @classmethod
    def xgcd(cls, poly1, poly2):
        """
        Performs the extended Euclidean algorithm on two given binary polynomials.
        """
        return xgcd(poly1, poly2, cls)

    @classmethod
    def gcd(cls, *poly_list):
        """
        Computes the greatest common divisor (gcd) of the arguments.
        """
        return functools.reduce(functools.partial(gcd, ring=cls), poly_list)

    @classmethod
    def lcm(cls, *poly_list):
        """
        Computes the least common multiple (lcm) of the arguments.
        """
        return functools.reduce(operator.mul, poly_list) // cls.gcd(*poly_list)


class BinaryPolynomialFraction:
    """
    Binary polynomial fraction. A *binary polynomial fraction* is a ratio of two binary polynomials (:class:`BinaryPolynomial`).
    """
    def __init__(self, numerator, denominator=0b1):
        self._numerator = BinaryPolynomial(numerator)
        self._denominator = BinaryPolynomial(denominator)
        if denominator == 0:
            raise ZeroDivisionError('Denominator cannot be zero')
        self._reduce_to_lowest_terms()

    def _reduce_to_lowest_terms(self):
        gcd = BinaryPolynomial.gcd(self._numerator, self._denominator)
        self._numerator //= gcd
        self._denominator //= gcd

    def __repr__(self):
        args = '{}, {}'.format(self._numerator, self._denominator)
        return '{}({})'.format(self.__class__.__name__, args)

    def __str__(self):
        if self._denominator == 0b1:
            return str(self._numerator)
        else:
            return str(self._numerator) + '/' + str(self._denominator)

    @property
    def numerator(self):
        """
        The numerator of the polynomial fraction.
        """
        return self._numerator

    @property
    def denominator(self):
        """
        The denominator of the polynomial fraction.
        """
        return self._denominator

    def __add__(self, other):
        numerator = self._numerator * other._denominator + self._denominator * other._numerator
        denominator = self._denominator * other._denominator
        return self.__class__(numerator, denominator)

    def __sub__(self, other):
        numerator = self._numerator * other._denominator - self._denominator * other._numerator
        denominator = self._denominator * other._denominator
        return self.__class__(numerator, denominator)

    def __mul__(self, other):
        numerator = self._numerator * other._numerator
        denominator = self._denominator * other._denominator
        return self.__class__(numerator, denominator)

    def __truediv__(self, other):
        numerator = self._numerator * other._denominator
        denominator = self._denominator * other._numerator
        return self.__class__(numerator, denominator)

    def __pow__(self, exponent):
        return power(self, exponent, self.__class__)

    def __eq__(self, other):
        return self._numerator * other._denominator == self._denominator * other._numerator

    def inverse(self):
        """
        Returns the multiplicative inverse the polynomial fraction.
        """
        return self.__class__(self._denominator, self._numerator)


class FiniteBifield:
    """
    Finite field with binary characteristic. Objects of this class represent a *finite field* :math:`\\mathrm{GF}(2^k)` (also known as *Galois field*), with *characteristic* :math:`2` and *degree* :math:`k`.  The constructor takes :math:`k` as a parameter.  Optionally, the *modulus*, or *primitive polynomial*, :math:`p(X)` may be specified; if not, the following default values will be chosen :cite:`Lin.Costello.04` (p. 42):

    ===========  =====================  ============  ============================
     :math:`k`    :math:`p(X)`           :math:`k`     :math:`p(X)`
    ===========  =====================  ============  ============================
     :math:`1`    :code:`0b11`           :math:`9`     :code:`0b1000010001`
     :math:`2`    :code:`0b111`          :math:`10`    :code:`0b10000001001`
     :math:`3`    :code:`0b1011`         :math:`11`    :code:`0b100000000101`
     :math:`4`    :code:`0b10011`        :math:`12`    :code:`0b1000001010011`
     :math:`5`    :code:`0b100101`       :math:`13`    :code:`0b10000000011011`
     :math:`6`    :code:`0b1000011`      :math:`14`    :code:`0b100010001000011`
     :math:`7`    :code:`0b10001001`     :math:`15`    :code:`0b1000000000000011`
     :math:`8`    :code:`0b100011101`    :math:`16`    :code:`0b10001000000001011`
    ===========  =====================  ============  ============================

    To construct *elements* of the finite field, call the finite field object. For example, :code:`field(0b1101)` will construct the element whose polynomial representation is :math:`X^3 + X^2 + 1`.

    .. rubric:: Examples

    >>> field = komm.FiniteBifield(4)
    >>> field
    FiniteBifield(4)
    >>> (field.characteristic, field.degree)
    (2, 4)
    >>> field.modulus
    BinaryPolynomial(0b10011)

    >>> field1 = komm.FiniteBifield(3, modulus=0b1011)
    >>> alpha1 = field1.primitive_element
    >>> [alpha1**i for i in range(7)]
    [0b1, 0b10, 0b100, 0b11, 0b110, 0b111, 0b101]
    >>> field2 = komm.FiniteBifield(3, modulus=0b1101)
    >>> alpha2 = field2.primitive_element
    >>> [alpha2**i for i in range(7)]
    [0b1, 0b10, 0b100, 0b101, 0b111, 0b11, 0b110]

    >>> field = komm.FiniteBifield(4)
    >>> x = field(0b1011)
    >>> y = field(0b1100)
    >>> x + y
    0b111
    >>> x * y
    0b1101
    >>> x / y
    0b10
    """
    def __init__(self, degree, modulus=None):
        """
        Constructor for the class. It expects the following parameters:

        :code:`degree` : :obj:`int`
            Degree :math:`k` of the finite field. Must be a positive integer.
        :code:`modulus` : :obj:`BinaryPolynomial` or :obj:`int`, optional
            Modulus (primitive polynomial) of the field, specified either as a :obj:`BinaryPolynomial` or as an :obj:`int` to be converted to the former. Must be an irreducible polynomial.
        """
        self._characteristic = 2
        self._degree = degree
        if modulus is None:
            PRIMITIVE_POLYNOMIALS = {
                1: 0b11,
                2: 0b111,
                3: 0b1011,
                4: 0b10011,
                5: 0b100101,
                6: 0b1000011,
                7: 0b10001001,
                8: 0b100011101,
                9: 0b1000010001,
                10: 0b10000001001,
                11: 0b100000000101,
                12: 0b1000001010011,
                13: 0b10000000011011,
                14: 0b100010001000011,
                15: 0b1000000000000011,
                16: 0b10001000000001011}
            self._modulus = BinaryPolynomial(PRIMITIVE_POLYNOMIALS[degree])
        else:
            self._modulus = BinaryPolynomial(modulus)

        assert self._modulus.degree == self._degree

    @property
    def characteristic(self):
        """
        The characteristic :math:`2` of the finite field. This property is read-only.
        """
        return self._characteristic

    @property
    def degree(self):
        """
        The degree :math:`k` of the finite field. This property is read-only.
        """
        return self._degree

    @property
    def modulus(self):
        """
        The modulus (primitive polynomial) :math:`p(X)` of the finite field. This property is read-only.
        """
        return self._modulus

    @property
    def order(self):
        """
        The order (number of elements) of the finite field. It is given by :math:`2^k`. This property is read-only.
        """
        return 2 ** self._degree

    @property
    def primitive_element(self):
        """
        A primitive element :math:`\\alpha` of the finite field. It satisfies :math:`p(\\alpha) = 0`, where :math:`p(X)` is the modulus (primitive polynomial) of the finite field. This property is read-only.
        """
        return self(2)

    # ~@functools.lru_cache(maxsize=None)
    def _multiply(self, x, y):
        return self((BinaryPolynomial(x) * BinaryPolynomial(y)) % self._modulus)

    # ~@functools.lru_cache(maxsize=None)
    def inverse(self, x):
        """
        Returns the multiplicative inverse of a given element.
        """
        d, s, _ = BinaryPolynomial.xgcd(BinaryPolynomial(x), self._modulus)
        if d._integer == 1:
            return self(s)
        else:
            raise ZeroDivisionError('This element does not have a multiplicative inverse')

    # ~@functools.lru_cache(maxsize=None)
    def logarithm(self, x, base=None):
        """
        Returns the logarithm of a given element, with respect to a given base.
        """
        if base is None:
            base = self.primitive_element
        for i in range(self.order):
            if base**i == x:
                return i
        return -1

    def power(self, x, exponent):
        """
        Returns a given power of a given element.
        """
        if exponent < 0:
            return power(self.inverse(x), -exponent, self)
        else:
            return power(x, exponent, self)

    @staticmethod
    def conjugates(x):
        """
        Returns the conjugates of a given element. See :cite:`Lin.Costello.04` (Sec. 2.5) for more details.
        """
        conjugate_list = []
        exponent = 0
        while True:
            y = x**(2**exponent)
            if y not in conjugate_list:
                conjugate_list.append(y)
            else:
                break
            exponent += 1
        return conjugate_list

    @staticmethod
    def minimal_polynomial(x):
        """
        Returns the minimal polynomial of a given element. See :cite:`Lin.Costello.04` (Sec. 2.5) fore more details.
        """
        one = x.field(1)
        monomials = [np.array([y, one], dtype=np.object) for y in x.conjugates()]
        coefficients = functools.reduce(np.convolve, monomials)
        return BinaryPolynomial.from_coefficients(int(c) for c in coefficients)

    def __repr__(self):
        args = '{}'.format(self._degree)
        return '{}({})'.format(self.__class__.__name__, args)

    def __call__(self, value):
        """
        Constructs elements of the finite field.
        """
        element = self._Element(value)
        element.field = self
        return element

    class _Element(int):
        """
        Elements of a FiniteBifield.

        Objects of this class represents *elements* of the finite field :math:`\\mathrm{GF}(2^k)`.
        """
        def __eq__(self, other): return int(self) == int(other) and self.field._modulus == other.field._modulus
        def __hash__(self): return hash((int(self), self.field._modulus))
        def __add__(self, other): return self.field(self ^ other)
        def __sub__(self, other): return self.field(self ^ other)
        def __mul__(self, other): return self.field._multiply(self, other)
        def inverse(self): return self.field.inverse(self)
        def __truediv__(self, other): return self * other.inverse()
        def logarithm(self, base=None): return self.field.logarithm(self, base)
        def __pow__(self, exponent): return self.field.power(self, exponent)
        def conjugates(self): return self.field.conjugates(self)
        def minimal_polynomial(self): return self.field.minimal_polynomial(self)
        def __repr__(self): return bin(self)
        def __str__(self): return bin(self)


class RationalPolynomial:
    """
    Rational polynomial. A *rational polynomial* is a polynomial whose coefficients are all rational numbers.

    .. rubric:: Examples

    >>> from fractions import Fraction
    >>> poly = komm.RationalPolynomial(['1/2', '0', '3'])  # 1/2 + 3 X^2
    >>> poly
    RationalPolynomial(['1/2', '0', '3'])
    """
    def __init__(self, coefficients):
        if isinstance(coefficients, (int, Fraction)):
            coefficients = [Fraction(coefficients)]
        elif isinstance(coefficients, self.__class__):
            coefficients = coefficients._coefficients

        self._coefficients = np.array(np.trim_zeros([Fraction(x) for x in coefficients], trim='b'))

    @classmethod
    def monomial(cls, degree, coefficient=1):
        """
        Constructs a monomial. This is an polynomial of the form :math:`cX^d`. It expects the following parameters:

        :code:`degree` : :obj:`int`
            The degree :math:`d` of the monomial.

        :code:`coefficient` : :obj:`int`, optional
            The coefficient :math:`c` of the monomial. The default value is :math:`1`.

        .. rubric:: Examples

        >>> komm.RationalPolynomial.monomial(4, 2)  # 2 X^4
        RationalPolynomial(['0', '0', '0', '0', '2'])
        """
        return cls([0] * degree + [coefficient])

    def coefficients(self, width=None):
        """
        Returns the coefficients of the polynomial.

        **Input:**

        :code:`width` : :obj:`int`, optional
            If this parameter is specified, the output will be filled with zeros on the right so that the its length will be the specified value.

        **Output:**

        :code:`coefficients` : 1D-array of :obj:`int`
            Coefficients of the polynomial. The :math:`i`-th element of the array stands for the coefficient of :math:`X^i`.

        .. rubric:: Examples

        >>> poly = komm.RationalPolynomial(['0', '1/3', '2/3'])  # (1/3) X + (2/3) X^2
        >>> poly.coefficients()
        array([Fraction(0, 1), Fraction(1, 3), Fraction(2, 3)], dtype=object)
        >>> poly.coefficients(width=5)
        array([Fraction(0, 1), Fraction(1, 3), Fraction(2, 3), Fraction(0, 1),
               Fraction(0, 1)], dtype=object)
        """
        if width is None:
            coefficients = self._coefficients
        else:
            coefficients = np.empty((width, ), dtype=np.object)
            coefficients[:self._coefficients.size] = self._coefficients
            coefficients[self._coefficients.size:] = Fraction(0)
        return coefficients

    @property
    def degree(self):
        """
        The degree of the polynomial. This property is read-only.

        .. rubric:: Examples

        >>> poly = komm.RationalPolynomial([1, 0, 3])  # 1 + 3X^2
        >>> poly.degree
        2
        """
        return self._coefficients.size - 1

    def __eq__(self, other):
        return np.array_equal(self._coefficients, other._coefficients)

    def __add__(self, other):
        if self.degree > other.degree:
            return self.__class__(self._coefficients + np.pad(other._coefficients, (0, self.degree - other.degree), mode='constant'))
        else:
            return self.__class__(np.pad(self._coefficients, (0, other.degree - self.degree), mode='constant') + other._coefficients)

    def __sub__(self, other):
        if self.degree > other.degree:
            return self.__class__(self._coefficients - np.pad(other._coefficients, (0, self.degree - other.degree), mode='constant'))
        else:
            return self.__class__(np.pad(self._coefficients, (0, other.degree - self.degree), mode='constant') - other._coefficients)

    def __neg__(self):
        return self.__class__(-self._coefficients)

    def __mul__(self, other):
        if self.degree == -1 or other.degree == -1:
            return self.__class__(0)
        return self.__class__(np.convolve(self._coefficients, other._coefficients))

    def __pow__(self, exponent):
        return power(self, exponent, self.__class__)

    def __divmod__(self, other):
        if other.degree == -1:
            raise ZeroDivisionError
        remainder = self._coefficients.tolist()
        quotient = [0] * (self.degree - other.degree + 1)
        for i in range(len(quotient)):
            quotient[-i - 1] = remainder[-1] / other._coefficients[-1]
            for j in range(1, len(other._coefficients)):
                remainder[-j - 1] -= quotient[-i - 1] * other._coefficients[-j - 1]
            del remainder[-1]
        return self.__class__(quotient), self.__class__(remainder)

    def __floordiv__(self, other):
        return divmod(self, other)[0]

    def __mod__(self, other):
        return divmod(self, other)[1]

    def evaluate(self, point):
        """
        Evaluates the polynomial at a given point. Uses Horner's method.

        **Input:**

        :code:`point` : ring-like type
            Any Python object supporting the operations of addition, subtraction, and multiplication.

        **Output:**

        :code:`result` : ring-like type
            The result of evaluating the binary polynomial at :code:`point`. It has the same type as :code:`point`.

        .. rubric:: Examples

        >>> poly = komm.RationalPolynomial([0, 1, 0, -1, 2])  # X - X^3 + 2 X^4
        >>> poly.evaluate(7)  # same as 7 - 7**3 + 2 * 7**4
        Fraction(4466, 1)
        >>> point = np.array([[1, 2], [3, 4]])
        >>> poly.evaluate(point)  # same as point - point**3 + 2 * point**4
        array([[Fraction(2, 1), Fraction(26, 1)],
               [Fraction(138, 1), Fraction(452, 1)]], dtype=object)
        """
        return horner(self, point)

    def __repr__(self):
        args = '{}'.format([str(f) for f in self._coefficients])
        return '{}({})'.format(self.__class__.__name__, args)

    @classmethod
    def gcd(cls, *poly_list):
        """
        Computes the greatest common divisor (gcd) of the arguments.
        """
        ans = functools.reduce(functools.partial(gcd, ring=cls), poly_list)
        a = np.lcm.reduce([coeff.denominator for coeff in ans._coefficients])
        ans *= cls([Fraction(a, 1)])
        b = np.gcd.reduce([coeff.numerator for coeff in ans._coefficients])
        ans *= cls([Fraction(1, b)])
        return ans

    @classmethod
    def lcm(cls, *poly_list):
        """
        Computes the least common multiple (lcm) of the arguments.
        """
        return functools.reduce(operator.mul, poly_list) // cls.gcd(*poly_list)


class RationalPolynomialFraction:
    """
    Rational polynomial fraction. A *rational polynomial fraction* is a ratio of two rational polynomials (:class:`RationalPolynomial`).
    """
    def __init__(self, numerator, denominator=1):
        self._numerator = RationalPolynomial(numerator)
        self._denominator = RationalPolynomial(denominator)
        if self._denominator.degree == -1:
            raise ZeroDivisionError('Denominator cannot be zero')
        self._reduce_to_lowest_terms()
        self._reduce_to_integer_coefficients()

    @classmethod
    def monomial(cls, degree, coefficient=1):
        return cls(RationalPolynomial.monomial(degree, coefficient))

    def _reduce_to_lowest_terms(self):
        gcd = RationalPolynomial.gcd(self._numerator, self._denominator)
        self._numerator //= gcd
        self._denominator //= gcd

    def _reduce_to_integer_coefficients(self):
        all_denominators = [x.denominator for x in self._numerator._coefficients] + \
                           [x.denominator for x in self._denominator._coefficients]
        a = np.lcm.reduce([n for n in all_denominators if n != 0])
        self._numerator *= RationalPolynomial([Fraction(a, 1)])
        self._denominator *= RationalPolynomial([Fraction(a, 1)])

        all_numerators = [x.numerator for x in self._numerator._coefficients] + \
                         [x.numerator for x in self._denominator._coefficients]
        b = np.gcd.reduce([n for n in all_numerators if n != 0])
        self._numerator *= RationalPolynomial([Fraction(1, b)])
        self._denominator *= RationalPolynomial([Fraction(1, b)])

    def __repr__(self):
        args = '{}, {}'.format(self._numerator, self._denominator)
        return '{}({})'.format(self.__class__.__name__, args)

    @property
    def numerator(self):
        """
        The numerator of the fraction.
        """
        return self._numerator

    @property
    def denominator(self):
        """
        The denominator of the fraction.
        """
        return self._denominator

    def __add__(self, other):
        numerator = self._numerator * other._denominator + self._denominator * other._numerator
        denominator = self._denominator * other._denominator
        return self.__class__(numerator, denominator)

    def __sub__(self, other):
        numerator = self._numerator * other._denominator - self._denominator * other._numerator
        denominator = self._denominator * other._denominator
        return self.__class__(numerator, denominator)

    def __neg__(self):
        return self.__class__(-self._numerator, self._denominator)

    def __mul__(self, other):
        numerator = self._numerator * other._numerator
        denominator = self._denominator * other._denominator
        return self.__class__(numerator, denominator)

    def __truediv__(self, other):
        numerator = self._numerator * other._denominator
        denominator = self._denominator * other._numerator
        return self.__class__(numerator, denominator)

    def __pow__(self, exponent):
        return power(self, exponent, self.__class__)

    def __eq__(self, other):
        return self._numerator * other._denominator == self._denominator * other._numerator

    def inverse(self):
        """
        Returns the multiplicative inverse the fraction.
        """
        return self.__class__(self._denominator, self._numerator)


def gcd(x, y, ring):
    """
    Performs the `Euclidean algorithm<https://en.wikipedia.org/wiki/Euclidean_algorithm>`_ with :code:`x` and :code:`y`.
    """
    if y == ring(0):
        return x
    else:
        return gcd(y, x % y, ring)


def xgcd(x, y, ring):
    """
    Performs the `extended Euclidean algorithm<https://en.wikipedia.org/wiki/Extended_Euclidean_algorithm>`_ with :code:`x` and :code:`y`.
    """
    if x == ring(0):
        return y, ring(0), ring(1)
    else:
        d, s, t = xgcd(y % x, x, ring)
        return d, t - s * (y // x), s


def power(x, n, ring):
    """
    Returns :code:`x**n` using the `exponentiation by squaring<https://en.wikipedia.org/wiki/Exponentiation_by_squaring>`_ algorithm.
    """
    if n == 0:
        return ring(1)
    elif n == 1:
        return x
    elif n % 2 == 0:
        return power(x * x, n // 2, ring)
    else:
        return x * power(x * x, n // 2, ring)


def binary_horner(poly, x):
    """
    Returns the binary polynomial :code:`poly` evaluated at point :code:`x`, using `Horner's method <https://en.wikipedia.org/wiki/Horner's_method>`_.  Any Python object supporting the operations of addition, subtraction, and multiplication may serve as the input point.
    """
    result = x - x  # zero
    for coefficient in reversed(poly.coefficients()):
        result *= x
        if coefficient:
            result += coefficient
    return result


def horner(poly, x):
    """
    Returns the polynomial :code:`poly` evaluated at point :code:`x`, using `Horner's method <https://en.wikipedia.org/wiki/Horner's_method>`_.  Any Python object supporting the operations of addition, subtraction, and multiplication may serve as the input point.
    """
    result = x - x  # zero
    for coefficient in reversed(poly.coefficients()):
        result = result * x + coefficient
    return result


def rref(M):
    """
    Computes the row-reduced echelon form of the matrix M modulo 2.

    Loosely based on
    [1] https://gist.github.com/rgov/1499136
    """
    M_rref = np.copy(M)
    n_rows, n_cols = M_rref.shape

    def pivot(row):
        f_list = np.flatnonzero(row)
        if f_list.size > 0:
            return f_list[0]
        else:
            return n_rows

    for r in range(n_rows):
        # Choose the pivot.
        possible_pivots = [pivot(row) for row in M_rref[r:]]
        p = np.argmin(possible_pivots) + r

        # Swap rows.
        M_rref[[r, p]] = M_rref[[p, r]]

        # Pivot column.
        f = pivot(M_rref[r])
        if f >= n_cols:
            continue

        # Subtract the row from others.
        for i in range(n_rows):
            if i != r and M_rref[i, f] != 0:
                M_rref[i] = (M_rref[i] + M_rref[r]) % 2

    return M_rref


# TODO: this should be the main function!
#       rref should call this instead
def xrref(M):
    """
    Computes the row-reduced echelon form of the matrix M modulo 2.

    Returns
    =======
    P

    M_rref

    pivots

    Such that :obj:`M_rref = P @ M` (where :obj:`@` stands for matrix multiplication).
    """
    eye = np.eye(M.shape[0], dtype=np.int)

    augmented_M = np.concatenate((np.copy(M), np.copy(eye)), axis=1)
    augmented_M_rref = rref(augmented_M)

    M_rref = augmented_M_rref[:, :M.shape[1]]
    P = augmented_M_rref[:, M.shape[1]:]

    pivots = []
    j = 0
    while len(pivots) < M_rref.shape[0] and j < M_rref.shape[1]:
        if np.array_equal(M_rref[:, j], eye[len(pivots)]):
            pivots.append(j)
        j += 1

    return P, M_rref, np.array(pivots)


def right_inverse(M):
    P, _, s_indices = xrref(M)
    M_rref_ri = np.zeros(M.T.shape, dtype=np.int)

    M_rref_ri[s_indices] = np.eye(len(s_indices), M.shape[0])
    M_ri = np.dot(M_rref_ri, P) % 2
    return M_ri


def null_matrix(M):
    (k, n) = M.shape
    _, M_rref, s_indices = xrref(M)
    N = np.empty((n - k, n), dtype=np.int)
    p_indices = np.setdiff1d(np.arange(M.shape[1]), s_indices)
    N[:, p_indices] = np.eye(n - k, dtype=np.int)
    N[:, s_indices] = M_rref[:, p_indices].T
    return N



import functools

import numpy as np

from ._algebra import \
    BinaryPolynomial

from ._util import \
    int2binlist, binlist2int

__all__ = ['BinarySequence',
           'BarkerSequence', 'WalshHadamardSequence',
           'LFSRSequence', 'GoldSequence', 'KasamiSequence',
           'ZadoffChuSequence']


class BinarySequence:
    """
    General binary sequence. It may be represented either in *bit format*, denoted by :math:`b[n]`, with elements in the set :math:`\\{ 0, 1 \\}`, or in *polar format*, denoted by :math:`a[n]`, with elements in the set :math:`\\{ \\pm 1 \\}`. The correspondences :math:`0 \\mapsto +1` and :math:`1 \\mapsto -1` from bit format to polar format is assumed.
    """
    def __init__(self, **kwargs):
        """
        Constructor for the class. It expects *exactly one* the following parameters:

        :code:`bit_sequence` : 1D-array of :obj:`int`
            The binary sequence in bit format. Must be an 1D-array with elements in :math:`\\{ 0, 1 \\}`.

        :code:`polar_sequence` : 1D-array of :obj:`int`
            The binary sequence in polar format. Must be an 1D-array with elements in :math:`\\{ \\pm 1 \\}`.
        """
        kwargs_set = set(kwargs.keys())
        if kwargs_set == {'bit_sequence'}:
            self._bit_sequence = np.array(kwargs['bit_sequence'], dtype=np.int)
            self._polar_sequence = (-1)**self._bit_sequence
            self._constructed_from = 'bit_sequence'
        elif kwargs_set == {'polar_sequence'}:
            self._polar_sequence = np.array(kwargs['polar_sequence'], dtype=np.int)
            self._bit_sequence = 1 * (self._polar_sequence < 0)
            self._constructed_from = 'polar_sequence'
        else:
            raise ValueError("Either specify 'bit_sequence' or 'polar_sequence'")

        self._length = self._bit_sequence.size

    def __repr__(self):
        args = 'bit_sequence={}'.format(self._bit_sequence.tolist())
        return '{}({})'.format(self.__class__.__name__, args)

    @property
    def bit_sequence(self):
        """
        The binary sequence in bit format, :math:`b[n] \\in \\{ 0, 1 \\}`.
        """
        return self._bit_sequence

    @property
    def polar_sequence(self):
        """
        The binary sequence in polar format, :math:`a[n] \\in \\{ \\pm 1 \\}`.
        """
        return self._polar_sequence

    @property
    def length(self):
        """
        The length (or period) :math:`L` of the binary sequence.
        """
        return self._length

    @property
    @functools.lru_cache()
    def _autocorrelation(self):
        seq = self._polar_sequence
        L = self._length
        return np.correlate(seq, seq, mode='full')[L - 1:]

    @property
    @functools.lru_cache()
    def _cyclic_autocorrelation(self):
        seq = self._polar_sequence
        L = self._length
        return np.array([np.dot(seq, np.roll(seq, ell)) for ell in range(L)])

    def autocorrelation(self, shifts=None, normalized=False):
        """
        Returns the autocorrelation :math:`R[\\ell]` of the binary sequence. This is defined as

        .. math::

            R[\\ell] = \\sum_{n \\in \\mathbb{Z}} a[n] a_{\\ell}[n],

        where :math:`a[n]` is the binary sequence in polar format, and :math:`a_{\\ell}[n] = a[n - \\ell]` is the sequence :math:`a[n]` shifted by :math:`\\ell` positions. The autocorrelation :math:`R[\\ell]` is even and satisfies :math:`R[\\ell] = 0` for :math:`|\\ell| \\geq L`, where :math:`L` is the length of the binary sequence.

        **Parameters:**

        :code:`shifts` : 1D-array of :obj:`int`, optional.
            An 1D array containing the values of :math:`\\ell` for which the autocorrelation will be computed. The default value is :code:`range(L)`, that is, :math:`[0 : L)`.

        :code:`normalized` : :obj:`boolean`, optional
            If :code:`True`, returns the autocorrelation divided by :math:`L`, where :math:`L` is the length of the binary sequence, so that :math:`R[0] = 1`. The default value is :code:`False`.
        """
        L = self._length
        shifts = np.arange(L) if shifts is None else np.array(shifts)
        autocorrelation = np.array([self._autocorrelation[abs(ell)] if abs(ell) < L else 0 for ell in shifts])
        if normalized:
            return autocorrelation / L
        else:
            return autocorrelation

    def cyclic_autocorrelation(self, shifts=None, normalized=False):
        """
        Returns the cyclic autocorrelation :math:`\\tilde{R}[\\ell]` of the binary sequence. This is defined as

        .. math::

            \\tilde{R}[\\ell] = \\sum_{n \\in [0:L)} a[n] \\tilde{a}_{\\ell}[n],

        where :math:`a[n]` is the binary sequence in polar format, and :math:`\\tilde{a}_{\\ell}[n]` is the sequence :math:`a[n]` cyclic-shifted by :math:`\\ell` positions. The cyclic autocorrelation :math:`\\tilde{R}[\\ell]` is even and periodic with period :math:`L`, where :math:`L` is the period of the binary sequence.

        **Parameters:**

        :code:`shifts` : 1D-array of :obj:`int`, optional.
            An 1D array containing the values of :math:`\\ell` for which the cyclic autocorrelation will be computed. The default value is :code:`range(L)`, that is, :math:`[0 : L)`.

        :code:`normalized` : :obj:`boolean`, optional
            If :code:`True`, returns the cyclic autocorrelation divided by :math:`L`, where :math:`L` is the length of the binary sequence, so that :math:`\\tilde{R}[0] = 1`. The default value is :code:`False`.
        """
        L = self._length
        shifts = np.arange(L) if shifts is None else np.array(shifts)
        cyclic_autocorrelation = self._cyclic_autocorrelation[shifts % L]
        if normalized:
            return cyclic_autocorrelation / L
        else:
            return cyclic_autocorrelation


class BarkerSequence(BinarySequence):
    """
    Barker sequence. A Barker sequence is a binary sequence (:obj:`BinarySequence`) with autocorrelation :math:`R[\\ell]` satisfying :math:`|R[\\ell]| \\leq 1`, for :math:`\\ell \\neq 0`. The only known Barker sequences (up to negation and reversion) are shown in the table below.

    ================  =============================
    Length :math:`L`  Barker sequence :math:`b[n]`
    ================  =============================
    :math:`2`         :math:`01` and :math:`00`
    :math:`3`         :math:`001`
    :math:`4`         :math:`0010` and :math:`0001`
    :math:`5`         :math:`00010`
    :math:`7`         :math:`0001101`
    :math:`11`        :math:`00011101101`
    :math:`13`        :math:`0000011001010`
    ================  =============================

    [1] https://en.wikipedia.org/wiki/Barker_code
    """
    def __init__(self, length):
        """
        Constructor for the class. It expects the following parameter:

        :code:`length` : :obj:`int`
            Length of the Barker sequence. Must be in the set :math:`\\{ 2, 3, 4, 5, 7, 11, 13 \\}`.

        .. rubric:: Examples

        >>> barker = komm.BarkerSequence(length=13)
        >>> barker.polar_sequence
        array([ 1,  1,  1,  1,  1, -1, -1,  1,  1, -1,  1, -1,  1])
        >>> barker.autocorrelation()
        array([13,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1])
        """
        super().__init__(bit_sequence=self._barker_sequence(length))

    def __repr__(self):
        args = 'length={}'.format(self.length)
        return '{}({})'.format(self.__class__.__name__, args)

    @staticmethod
    def _barker_sequence(length):
        DICT = {
            2: [0, 1],
            3: [0, 0, 1],
            4: [0, 0, 1, 0],
            5: [0, 0, 0, 1, 0],
            7: [0, 0, 0, 1, 1, 0, 1],
            11: [0, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1],
            13: [0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 1, 0],
        }
        return DICT[length]


class WalshHadamardSequence(BinarySequence):
    """
    Walsh--Hadamard sequence. Consider the following recursive matrix construction:

    .. math::

        H_{1} =
        \\begin{bmatrix}
            +1
        \\end{bmatrix}, \\qquad
        H_{2^n} =
        \\begin{bmatrix}
            H_{2^{n-1}} & H_{2^{n-1}} \\\\
            H_{2^{n-1}} & -H_{2^{n-1}}
        \\end{bmatrix},

    for :math:`n = 1, 2, \\ldots`. For example, for :math:`n = 3`,

    .. math::

        H_{8} =
        \\begin{bmatrix}
            +1 & +1 & +1 & +1 & +1 & +1 & +1 & +1 \\\\
            +1 & -1 & +1 & -1 & +1 & -1 & +1 & -1 \\\\
            +1 & +1 & -1 & -1 & +1 & +1 & -1 & -1 \\\\
            +1 & -1 & -1 & +1 & +1 & -1 & -1 & +1 \\\\
            +1 & +1 & +1 & +1 & -1 & -1 & -1 & -1 \\\\
            +1 & -1 & +1 & -1 & -1 & +1 & -1 & +1 \\\\
            +1 & +1 & -1 & -1 & -1 & -1 & +1 & +1 \\\\
            +1 & -1 & -1 & +1 & -1 & +1 & +1 & -1 \\\\
        \\end{bmatrix}

    The above matrix is said to be in *natural ordering*. If the rows of the matrix are rearranged by first applying the bit-reversal permutation and then the Gray-code permutation, the following matrix is obtained:

    .. math::
        H_{8}^{\\mathrm{s}} =
        \\begin{bmatrix}
            +1 & +1 & +1 & +1 & +1 & +1 & +1 & +1 \\\\
            +1 & +1 & +1 & +1 & -1 & -1 & -1 & -1 \\\\
            +1 & +1 & -1 & -1 & -1 & -1 & +1 & +1 \\\\
            +1 & +1 & -1 & -1 & +1 & +1 & -1 & -1 \\\\
            +1 & -1 & -1 & +1 & +1 & -1 & -1 & +1 \\\\
            +1 & -1 & -1 & +1 & -1 & +1 & +1 & -1 \\\\
            +1 & -1 & +1 & -1 & -1 & +1 & -1 & +1 \\\\
            +1 & -1 & +1 & -1 & +1 & -1 & +1 & -1 \\\\
        \\end{bmatrix}

    The above matrix is said to be in *sequency ordering*. It has the property that row :math:`i` has exactly :math:`i` signal changes.

    The Walsh--Hadamard sequence of *length* :math:`L` and *index* :math:`i \\in [0 : L)` is the binary sequence (:obj:`BinarySequence`) whose polar format is the :math:`i`-th row of :math:`H_L`, if assuming natural ordering, or :math:`H_L^{\\mathrm{s}}`, if assuming sequency ordering.

    [1] https://en.wikipedia.org/wiki/Hadamard_matrix; [2] https://en.wikipedia.org/wiki/Walsh_matrix
    """
    def __init__(self, length, ordering='natural', index=0):
        """
        Constructor for the class. It expects the following parameters:

        :code:`length` : :obj:`int`
            Length :math:`L` of the Walsh--Hadamard sequence. Must be a power of two.

        :code:`ordering` : :obj:`str`, optional
            Ordering to be assumed. Should be one of :code:`'natural'`, :code:`'sequency'`, or :code:`'dyadic'`. The default value is :code:`'natural'`.

        :code:`index` : :obj:`int`, optional
            Index of the Walsh--Hadamard sequence, with respect to the ordering assumed. Must be in the set :math:`[0 : L)`. The default value is :code:`0`.

        .. rubric:: Examples

        >>> walsh_hadamard = komm.WalshHadamardSequence(length=64, ordering='sequency', index=60)
        >>> walsh_hadamard.polar_sequence[:16]
        array([ 1, -1,  1, -1,  1, -1,  1, -1,  1, -1,  1, -1,  1, -1,  1, -1])

        >>> walsh_hadamard = komm.WalshHadamardSequence(length=128, ordering='natural', index=60)
        >>> walsh_hadamard.polar_sequence[:16]
        array([ 1,  1,  1,  1, -1, -1, -1, -1, -1, -1, -1, -1,  1,  1,  1,  1])
        """
        if length & (length - 1):
            raise ValueError("The length of sequence must be a power of two")

        if not 0 <= index < length:
            raise ValueError("Parameter 'index' must be in [0, ..., length)")

        if ordering == 'natural':
            natural_index = index
        elif ordering == 'sequency':
            width = (length - 1).bit_length()
            index_gray = index ^ (index >> 1)
            natural_index = binlist2int(reversed(int2binlist(index_gray, width)))
        elif ordering == 'dyadic':
            raise NotImplementedError
        else:
            raise ValueError("Parameter 'ordering' must be 'natural', 'sequency' or 'dyadic'")

        self._index = index
        self._ordering = ordering
        super().__init__(polar_sequence=self._hadamard_matrix(length)[natural_index])

    def __repr__(self):
        args = "length={}, ordering='{}', index={}".format(self._length, self._ordering, self._index)
        return '{}({})'.format(self.__class__.__name__, args)

    @property
    def index(self):
        """
        The index of the Walsh--Hadamard sequence, with respect to the ordering assumed.
        """
        return self._index

    @property
    def ordering(self):
        """
        The ordering assumed.
        """
        return self._ordering

    @staticmethod
    def _hadamard_matrix(length):
        h = np.array([[1]])
        g = np.array([[1, 1], [1, -1]])
        for _ in range(length.bit_length() - 1):
            h = np.kron(h, g)
        return h


class LFSRSequence(BinarySequence):
    """
    Linear-feedback shift register (LFSR) sequence. It is the binary sequence (:obj:`BinarySequence`) obtained from the output of a LFSR. The LFSR feedback taps are specified as a binary polynomial :math:`p(X)` of degree :math:`n`, called the *feedback polynomial*. More specifically: if bit :math:`i` of the LFSR is tapped, for :math:`i \\in [1 : n]`, then the coefficient of :math:`X^i` in :math:`p(X)` is :math:`1`; otherwise, it is :math:`0`; moreover, the coefficient of :math:`X^0` in :math:`p(X)` is always :math:`1`. For example, the feedback polynomial corresponding to the LFSR in the figure below is :math:`p(X) = X^5 + X^2 + 1`, whose integer representation is :code:`0b100101`.

    .. image:: figures/lfsr_5_2.png
       :alt: Linear-feedback shift register example.
       :align: center

    The start state of the machine is specified by the so called *start state polynomial*. More specifically, the coefficient of :math:`X^i` in the start state polynomial is equal to the initial value of bit :math:`i` of the LFSR.

    .. rubric:: Maximum-length sequences

    If the feedback polynomial :math:`p(X)` is primitive, then the corresponding LFSR sequence will be a *maximum-length sequence* (MLS). Such sequences have the following cyclic autocorrelation:

    .. math::

        R[\\ell] =
        \\begin{cases}
            L, & \\ell = 0, \\pm L, \\pm 2L, \\ldots, \\\\
            -1, & \\text{otherwise},
        \\end{cases}

    where :math:`L` is the length of the sequence. The constructor :func:`maximum_length_sequence` can be use to construct an MLS.

    [1] https://en.wikipedia.org/wiki/Linear-feedback_shift_register; [2] https://en.wikipedia.org/wiki/Maximum_length_sequence

    .. rubric:: Examples

    >>> lfsr = komm.LFSRSequence(feedback_polynomial=0b100101)
    >>> lfsr.bit_sequence  #doctest: +NORMALIZE_WHITESPACE
    array([0, 0, 0, 0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1])
    >>> lfsr.cyclic_autocorrelation()  #doctest: +NORMALIZE_WHITESPACE
    array([31, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])
    """
    def __init__(self, feedback_polynomial, start_state_polynomial=0b1):
        """
        Constructor for the class. It expects the following parameters:

        :code:`feedback_polynomial` : :obj:`BinaryPolynomial` or :obj:`int`
            The feedback polynomial of the LFSR, specified either as a :obj:`BinaryPolynomial` or as an :obj:`int` to be converted to the former.

        :code:`start_state_polynomial` : :obj:`BinaryPolynomial` or :obj:`int`, optional.
            The start state polynomial of the LFSR, specified either as a :obj:`BinaryPolynomial` or as an :obj:`int` to be converted to the former. The default value is :code:`0b1`.
        """
        self._feedback_polynomial = BinaryPolynomial(feedback_polynomial)
        self._start_state_polynomial = BinaryPolynomial(start_state_polynomial)
        super().__init__(bit_sequence=self._lfsr_sequence(self._feedback_polynomial, self._start_state_polynomial))

    @classmethod
    def maximum_length_sequence(cls, degree):
        """
        Constructs a maximum-length sequences (MLS) of a given degree. It expects the following parameter:

        :code:`degree` : :obj:`int`
            The degree :math:`n` of the MLS. Only degrees in the range :math:`[1 : 16]` are implemented.

        The feedback polynomial :math:`p(X)` is chosen according to the following table of primitive polynomials.

        ================  ================================  ================  ================================
        Degree :math:`n`  Feedback polynomial :math:`p(X)`  Degree :math:`n`  Feedback polynomial :math:`p(X)`
        ================  ================================  ================  ================================
        :math:`1`         :code:`0b11`                      :math:`9`         :code:`0b1000010001`
        :math:`2`         :code:`0b111`                     :math:`10`        :code:`0b10000001001`
        :math:`3`         :code:`0b1011`                    :math:`11`        :code:`0b100000000101`
        :math:`4`         :code:`0b10011`                   :math:`12`        :code:`0b1000001010011`
        :math:`5`         :code:`0b100101`                  :math:`13`        :code:`0b10000000011011`
        :math:`6`         :code:`0b1000011`                 :math:`14`        :code:`0b100010001000011`
        :math:`7`         :code:`0b10001001`                :math:`15`        :code:`0b1000000000000011`
        :math:`8`         :code:`0b100011101`               :math:`16`        :code:`0b10001000000001011`
        ================  ================================  ================  ================================
        """
        PRIMITIVE_POLYNOMIALS = {
            1: 0b11,
            2: 0b111,
            3: 0b1011,
            4: 0b10011,
            5: 0b100101,
            6: 0b1000011,
            7: 0b10001001,
            8: 0b100011101,
            9: 0b1000010001,
            10: 0b10000001001,
            11: 0b100000000101,
            12: 0b1000001010011,
            13: 0b10000000011011,
            14: 0b100010001000011,
            15: 0b1000000000000011,
            16: 0b10001000000001011}
        return cls(PRIMITIVE_POLYNOMIALS[degree])

    def __repr__(self):
        args = 'feedback_polynomial={}'.format(self._feedback_polynomial)
        if self._start_state_polynomial != BinaryPolynomial(0b1):
            args += ', start_state_polynomial={}'.format(self._start_state_polynomial)
        return '{}({})'.format(self.__class__.__name__, args)

    @property
    def feedback_polynomial(self):
        """
        The feedback polynomial :math:`p(X)` of the LFSR.
        """
        return self._feedback_polynomial

    @property
    def start_state_polynomial(self):
        """
        The start state polynomial of the LFSR.
        """
        return self._start_state_polynomial

    @staticmethod
    def _lfsr_sequence(feedback_polynomial, start_state_polynomial):
        taps = (feedback_polynomial + BinaryPolynomial(1)).exponents()
        start_state = start_state_polynomial.coefficients(width=feedback_polynomial.degree)
        m = taps[-1]
        L = 2**m - 1
        state = np.copy(start_state)
        code = np.empty(L, dtype=np.int)
        for i in range(L):
            code[i] = state[-1]
            state[-1] = np.count_nonzero(state[taps - 1]) % 2
            state = np.roll(state, 1)
        return code


class GoldSequence:
    """
    Gold sequence [Not implemented yet].
    """
    pass


class KasamiSequence:
    """
    Kasami sequence [Not implemented yet].
    """
    pass


class ZadoffChuSequence:
    """
    Zadoff–Chu sequence [Not implemented yet].
    """
    pass



import numpy as np

from ._util import \
    _entropy, _mutual_information

__all__ = ['AWGNChannel',
           'DiscreteMemorylessChannel',
           'BinarySymmetricChannel', 'BinaryErasureChannel']


class AWGNChannel:
    """
    Additive white gaussian noise (AWGN) channel. It is defined by

    .. math::

        Y_n = X_n + Z_n,

    where :math:`X_n` is the channel *input signal*, :math:`Y_n` is the channel *output signal*, and :math:`Z_n` is the *noise*, which is :term:`i.i.d.` according to a gaussian distribution with zero mean. The channel *signal-to-noise ratio* is calculated by

    .. math::

        \\mathrm{SNR} = \\frac{P}{N},

    where :math:`P = \\mathrm{E}[X^2_n]` is the average power of the input signal, and :math:`N = \\mathrm{E}[Z^2_n]` is the average power (and variance) of the noise.

    References: :cite:`Cover.Thomas.06` (Ch. 9)

    To invoke the channel, call the object giving the input signal as parameter (see example below).
    """
    def __init__(self, snr=np.inf, signal_power=1.0):
        """Constructor for the class. It expects the following parameters:

        :code:`snr` : :obj:`float`, optional
            The channel signal-to-noise ratio :math:`\\mathrm{SNR}` (linear, not decibel). The default value is :code:`np.inf`.

        :code:`signal_power` : :obj:`float` or :obj:`str`, optional
            The input signal power :math:`P`.  If equal to the string :code:`'measured'`, then every time the channel is invoked the input signal power will be computed from the input itself (i.e., its squared Euclidean norm). The default value is :code:`1.0`.

        .. rubric:: Examples

        >>> awgn = komm.AWGNChannel(snr=100.0, signal_power=1.0)
        >>> x = [1.0, 3.0, -3.0, -1.0, -1.0, 1.0, 3.0, 1.0, -1.0, 3.0]
        >>> y = awgn(x); y  #doctest:+SKIP
        array([ 1.10051445,  3.01308154, -2.97997111, -1.1229903 , -0.90890299,
                1.12650432,  2.88952462,  0.99352172, -1.2072787 ,  3.27131731])
        """
        self.snr = snr
        self.signal_power = signal_power

    @property
    def snr(self):
        """
        The signal-to-noise ratio :math:`\\mathrm{SNR}` (linear, not decibel) of the channel. This is a read-and-write property.
        """
        return self._snr

    @snr.setter
    def snr(self, value):
        self._snr = float(value)

    @property
    def signal_power(self):
        """
        The input signal power :math:`P`. This is a read-and-write property.
        """
        return self._signal_power

    @signal_power.setter
    def signal_power(self, value):
        self._signal_power = float(value)

    def capacity(self):
        """
        Returns the channel capacity :math:`C`. It is given by :math:`C = \\frac{1}{2}\\log_2(1 + \\mathrm{SNR})`, in bits per dimension.

        .. rubric:: Examples

        >>> awgn = komm.AWGNChannel(snr=63.0)
        >>> awgn.capacity()
        3.0
        """
        return 0.5 * np.log1p(self._snr) / np.log(2.0)

    def __call__(self, input_signal):
        input_signal = np.array(input_signal)
        size = input_signal.size

        if self._signal_power == 'measured':
            signal_power = np.linalg.norm(input_signal)**2 / size
        else:
            signal_power = self._signal_power

        noise_power = signal_power / self._snr

        if input_signal.dtype == np.complex:
            noise = np.sqrt(noise_power / 2) * (np.random.normal(size=size) + 1j * np.random.normal(size=size))
        else:
            noise = np.sqrt(noise_power) * np.random.normal(size=size)

        return input_signal + noise

    def __repr__(self):
        args = 'snr={}, signal_power={}'.format(self._snr, self._signal_power)
        return '{}({})'.format(self.__class__.__name__, args)


class DiscreteMemorylessChannel:
    """
    Discrete memoryless channel (DMC). It is defined by an *input alphabet* :math:`\\mathcal{X}`, an *output alphabet* :math:`\\mathcal{Y}`, and a *transition probability matrix* :math:`p_{Y \\mid X}`. Here, for simplicity, the input and output alphabets are always taken as :math:`\\mathcal{X} = \\{ 0, 1, \\ldots, |\\mathcal{X}| - 1 \\}` and :math:`\\mathcal{Y} = \\{ 0, 1, \\ldots, |\\mathcal{Y}| - 1 \\}`, respectively. The transition probability matrix :math:`p_{Y \\mid X}`, of size :math:`|\\mathcal{X}|`-by-:math:`|\\mathcal{Y}|`, gives the conditional probability of receiving :math:`Y = y` given that :math:`X = x` is transmitted.

    References: :cite:`Cover.Thomas.06` (Ch. 7)

    To invoke the channel, call the object giving the input signal as parameter (see example below).
    """
    def __init__(self, transition_matrix):
        """
        Constructor for the class. It expects the following parameter:

        :code:`transition_matrix` : 2D-array of :obj:`float`
            The channel transition probability matrix :math:`p_{Y \\mid X}`. The element in row :math:`x \\in \\mathcal{X}` and column :math:`y \\in \\mathcal{Y}` must be equal to :math:`p_{Y \\mid X}(y \\mid x)`.

        .. rubric:: Examples

        >>> dmc = komm.DiscreteMemorylessChannel([[0.9, 0.05, 0.05], [0.0, 0.5, 0.5]])
        >>> x = [0, 1, 0, 1, 1, 1, 0, 0, 0, 1]
        >>> y = dmc(x); y  #doctest:+SKIP
        array([0, 2, 0, 2, 1, 1, 0, 0, 0, 2])
        """
        self.transition_matrix = transition_matrix
        self._arimoto_blahut_kwargs = {'max_iters': 1000, 'error_tolerance': 1e-12}

    @property
    def transition_matrix(self):
        """
        The channel transition probability matrix :math:`p_{Y \\mid X}`. This is a read-and-write property.
        """
        return self._transition_matrix

    @transition_matrix.setter
    def transition_matrix(self, value):
        self._transition_matrix = np.array(value, dtype=np.float)
        self._input_cardinality, self._output_cardinality = self._transition_matrix.shape

    @property
    def input_cardinality(self):
        """
        The channel input cardinality :math:`|\\mathcal{X}|`. This property is read-only.
        """
        return self._input_cardinality

    @property
    def output_cardinality(self):
        """
        The channel output cardinality :math:`|\\mathcal{Y}|`. This property is read-only.
        """
        return self._output_cardinality

    def mutual_information(self, input_pmf, base=2.0):
        """
        Computes the mutual information :math:`\\mathrm{I}(X ; Y)` between the input :math:`X` and the output :math:`Y` of the channel. It is given by

        .. math::

           \\mathrm{I}(X ; Y) = \\mathrm{H}(X) - \\mathrm{H}(X \\mid Y),

        where :math:`\\mathrm{H}(X)` is the the entropy of :math:`X` and :math:`\\mathrm{H}(X \\mid Y)` is the conditional entropy of :math:`X` given :math:`Y`. By default, the base of the logarithm is :math:`2`, in which case the mutual information is measured in bits.

        References: :cite:`Cover.Thomas.06` (Ch. 2)

        **Input:**

        :code:`input_pmf` : 1D-array of :obj:`float`
            The probability mass function :math:`p_X` of the channel input :math:`X`. It must be a valid :term:`pmf`, that is, all of its values must be non-negative and sum up to :math:`1`.

        :code:`base` : :obj:`float` or :obj:`str`, optional
            The base of the logarithm to be used. It must be a positive float or the string :code:`'e'`. The default value is :code:`2.0`.

        **Output:**

        :code:`mutual_information` : :obj:`float`
            The mutual information :math:`\\mathrm{I}(X ; Y)` between the input :math:`X` and the output :math:`Y`.

        .. rubric:: Examples

        >>> dmc = komm.DiscreteMemorylessChannel([[0.6, 0.3, 0.1], [0.7, 0.1, 0.2], [0.5, 0.05, 0.45]])
        >>> dmc.mutual_information([1/3, 1/3, 1/3])
        0.12381109879798724
        """
        return _mutual_information(input_pmf, self._transition_matrix, base)


    def capacity(self):
        """
        Returns the channel capacity :math:`C`. It is given by :math:`C = \\max_{p_X} \\mathrm{I}(X;Y)`. This method computes the channel capacity via the Arimoto--Blahut algorithm. See :cite:`Cover.Thomas.06` (Sec. 10.8).

        .. rubric:: Examples

        >>> dmc = komm.DiscreteMemorylessChannel([[0.6, 0.3, 0.1], [0.7, 0.1, 0.2], [0.5, 0.05, 0.45]])
        >>> dmc.capacity()
        0.1616318609548566
        """
        initial_guess = np.ones(self._input_cardinality, dtype=np.float) / self._input_cardinality
        optimal_input_pmf = self._arimoto_blahut(self._transition_matrix, initial_guess, **self._arimoto_blahut_kwargs)
        return _mutual_information(optimal_input_pmf, self._transition_matrix)

    def __call__(self, input_sequence):
        output_sequence = [np.random.choice(self._output_cardinality, p=self._transition_matrix[input_symbol])
                           for input_symbol in input_sequence]
        return np.array(output_sequence)

    def __repr__(self):
        args = 'transition_matrix={}'.format(self._transition_matrix.tolist())
        return '{}({})'.format(self.__class__.__name__, args)

    @staticmethod
    def _arimoto_blahut(transition_matrix, initial_guess, max_iters, error_tolerance):
        """
        Arimoto--Blahut algorithm for channel capacity. See :cite:`Cover.Thomas.06` (Sec. 10.8).
        """
        p = transition_matrix
        r = initial_guess
        last_r = np.full_like(r, fill_value=np.inf)
        iters = 0
        while iters < max_iters and np.amax(np.abs(r - last_r)) > error_tolerance:
            last_r = r
            q = r[np.newaxis].T * p
            q /= np.sum(q, axis=0)
            r = np.product(q ** p, axis=1)
            r /= np.sum(r, axis=0)
            iters += 1
        return r


class BinarySymmetricChannel(DiscreteMemorylessChannel):
    """
    Binary symmetric channel (BSC). It is a discrete memoryless channel (:obj:`DiscreteMemorylessChannel`) with input and output alphabets given by :math:`\\mathcal{X} = \\mathcal{Y} = \\{ 0, 1 \\}`, and transition probability matrix given by

    .. math::

        p_{Y \\mid X} = \\begin{bmatrix} 1-p & p \\\\ p & 1-p \\end{bmatrix},

    where the parameter :math:`p` is called the *crossover probability* of the channel. Equivalently, a BSC with crossover probability :math:`p` may be defined by

    .. math::
        Y_n = X_n + Z_n,

    where :math:`Z_n` are :term:`i.i.d.` Bernouli random variables with :math:`\\Pr[Z_n = 1] = p`.

    References: :cite:`Cover.Thomas.06` (Sec. 7.1.4)

    To invoke the channel, call the object giving the input signal as parameter (see example below).
    """
    def __init__(self, crossover_probability=0.0):
        """
        Constructor for the class. It expects the following parameter:

        :code:`crossover_probability` : :obj:`float`, optional
            The channel crossover probability :math:`p`. Must satisfy :math:`0 \\leq p \\leq 1`. The default value is :code:`0.0`.

        .. rubric:: Examples

        >>> bsc = komm.BinarySymmetricChannel(0.1)
        >>> x = [0, 1, 1, 1, 0, 0, 0, 0, 0, 1]
        >>> y = bsc(x); y  #doctest:+SKIP
        array([0, 1, 1, 1, 0, 0, 0, 1, 0, 0])
        """
        self.crossover_probability = crossover_probability

    @property
    def crossover_probability(self):
        """
        The crossover probability :math:`p` of the channel. This is a read-and-write property.
        """
        return self._crossover_probability

    @crossover_probability.setter
    def crossover_probability(self, value):
        self._crossover_probability = p = float(value)
        self.transition_matrix = np.array([[1 - p, p], [p, 1 - p]])

    def capacity(self):
        """
        Returns the channel capacity :math:`C`. It is given by :math:`C = 1 - \\mathcal{H}(p)`. See :cite:`Cover.Thomas.06` (Sec. 7.1.4).

        .. rubric:: Examples

        >>> bsc = komm.BinarySymmetricChannel(0.25)
        >>> bsc.capacity()
        0.18872187554086717
        """
        return 1.0 - _entropy(np.array([self._crossover_probability, 1.0 - self._crossover_probability]))

    def __call__(self, input_sequence):
        error_pattern = (np.random.rand(np.size(input_sequence)) < self._crossover_probability).astype(np.int)
        return (input_sequence + error_pattern) % 2

    def __repr__(self):
        args = 'crossover_probability={}'.format(self._crossover_probability)
        return '{}({})'.format(self.__class__.__name__, args)


class BinaryErasureChannel(DiscreteMemorylessChannel):
    """
    Binary erasure channel (BEC). It is a discrete memoryless channel (:obj:`DiscreteMemorylessChannel`) with input alphabet :math:`\\mathcal{X} = \\{ 0, 1 \\}`, output alphabet :math:`\\mathcal{Y} = \\{ 0, 1, 2 \\}`, and transition probability matrix given by

    .. math::

        p_{Y \\mid X} =
        \\begin{bmatrix}
            1 - \\epsilon & 0 & \\epsilon \\\\
            0 & 1 - \\epsilon & \\epsilon
        \\end{bmatrix},

    where the parameter :math:`\\epsilon` is called the *erasure probability* of the channel.

    References: :cite:`Cover.Thomas.06` (Sec. 7.1.5)

    To invoke the channel, call the object giving the input signal as parameter (see example below).
    """
    def __init__(self, erasure_probability=0.0):
        """
        Constructor for the class. It expects the following parameter:

        :code:`erasure_probability` : :obj:`float`, optional
            The channel erasure probability :math:`\\epsilon`. Must satisfy :math:`0 \\leq \\epsilon \\leq 1`. Default value is :code:`0.0`.

        .. rubric:: Examples

        >>> bec = komm.BinaryErasureChannel(0.1)
        >>> x = [1, 1, 1, 0, 0, 0, 1, 0, 1, 0]
        >>> y = bec(x); y  #doctest:+SKIP
        array([1, 1, 1, 2, 0, 0, 1, 0, 1, 0])
        """
        self.erasure_probability = erasure_probability

    @property
    def erasure_probability(self):
        """
        The erasure probability :math:`\\epsilon` of the channel. This is a read-and-write property.
        """
        return self._erasure_probability

    @erasure_probability.setter
    def erasure_probability(self, value):
        self._erasure_probability = e = float(value)
        self.transition_matrix  = [[1 - e, 0, e], [0, 1 - e, e]]

    def capacity(self):
        """
        Returns the channel capacity :math:`C`. It is given by :math:`C = 1 - \\epsilon`.  See :cite:`Cover.Thomas.06` (Sec. 7.1.5).

        .. rubric:: Examples

        >>> bec = komm.BinaryErasureChannel(0.25)
        >>> bec.capacity()
        0.75
        """
        return 1.0 - self._erasure_probability

    def __call__(self, input_sequence):
        erasure_pattern = (np.random.rand(np.size(input_sequence)) < self._erasure_probability)
        output_sequence = np.copy(input_sequence)
        output_sequence[erasure_pattern] = 2
        return output_sequence

    def __repr__(self):
        args = 'erasure_probability={}'.format(self._erasure_probability)
        return '{}({})'.format(self.__class__.__name__, args)
